import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os
import argparse
import math
import matplotlib.patches as patches

plt.rcParams.update({
    "font.family": "serif",
    "text.usetex": True,
    'text.latex.preamble': "\\usepackage{libertine}\\usepackage[libertine]{newtxmath}",
})

### Global Definitions
color_mapping = {
    'NTA': '#778da9',
    'T2': '#e07a5f',
    'T1': '#81b29a',
    'T0': '#f2cc8f',
    'Full': '#39568C',
    'Full': '#39568C',
    'Adjacent': '#73D055',
    'Sequential': '#588157',
    'Random': '#c1121f'
}

fill_mapping = {
    'cycles-no-stall': '',
    'cycles-mem-stall': '//',
    'cycles-other-stall': '...'
}

prefetch_locality_names = {0: 'NTA', 1: 'T2', 2: 'T1', 3: 'T0'}

'''
Multiplot with prefetch distance performance for each block size.
'''
def plot_prefetch_distances(data, file_path_out: str, metric: str, y_axis_label: str, hyper_threads: int, load_heavy: int, is_interleaved: bool, page_size: str):
    if 'hyper-threads' in data.columns:
        data = data.query(f'`hyper-threads` == {hyper_threads}')

    if 'page-size' in data.columns:
        data = data.query(f'`page-size` == "{page_size}"')

    if 'is-interleaved-prefetching' in data.columns:
        # Filter the data based on the specified conditions
        full_prefetch_data = data[(data['prefetch-distance'] > 0) &
             (data['is-random'] == 1) &
             (data['cache-lines'] == data['prefetched-cache-lines']) &
             (data['is-adjacent-pref'] == 0) &
             (data['is-interleaved-prefetching'] == int(is_interleaved))
         ]
    else:
        # Filter the data based on the specified conditions
        full_prefetch_data = data[(data['prefetch-distance'] > 0) &
              (data['is-random'] == 1) &
              (data['cache-lines'] == data['prefetched-cache-lines']) &
              (data['is-adjacent-pref'] == 0)
          ]
    
    if 'load-heavy' in data.columns:
        full_prefetch_data = full_prefetch_data.query(f'`load-heavy` == {load_heavy}')        
    
    max_y_value = data[metric].max()
    unique_cache_lines = full_prefetch_data['cache-lines'].unique()
    unique_prefetch_localities = full_prefetch_data['prefetch-locality'].unique()
    unique_prefetch_distances = sorted(full_prefetch_data['prefetch-distance'].unique())

    num_rows = int(math.ceil(len(unique_cache_lines) / 3))
    
    plt.figure(figsize=(15, 12))
    for i, cache_line in enumerate(unique_cache_lines, 1):
        # Limit to a 3x3 grid
        #if i > 9:
        #    break

        ax = plt.subplot(num_rows, 3, i)
        cache_line_data = full_prefetch_data[full_prefetch_data['cache-lines'] == cache_line]
        
        for locality in unique_prefetch_localities:
            subset_data = cache_line_data[cache_line_data['prefetch-locality'] == locality]
            ax.plot(
                subset_data['prefetch-distance'], 
                subset_data[metric], 
                label=f'{prefetch_locality_names[locality]}',
                color=color_mapping[prefetch_locality_names[locality]],
                marker='o', 
                linestyle='-'
            ) 
            
        
        # Add horizontal (base-) line for random and sequential
        sequential_data = data[(data['is-random'] == 0) & (data['cache-lines'] == cache_line) & (data['prefetch-locality'] == -1)].reset_index()
        if len(sequential_data) > 0:
            ax.axhline(sequential_data.at[0, metric], color=color_mapping['Sequential'], linestyle='-', label='Sequential')
            
        random_data = data[(data['is-random'] == 1) & (data['prefetch-distance'] == 0) & (data['cache-lines'] == cache_line)].reset_index()
        if len(random_data) > 0:
            ax.axhline(random_data.at[0, metric], color=color_mapping['Random'], linestyle='-', label='Random')
        
        # Set x-ticks for prefetch-distance
        ax.set_xticks(unique_prefetch_distances)
        ax.set_xticklabels(unique_prefetch_distances)
        
        ax.set_title(f'{cache_line} contiguous cache lines' if cache_line > 1 else '1 cache line', fontsize=11)
        ax.set_xlabel('Prefetch Distance')
        ax.set_ylabel(y_axis_label)
        ax.set_ylim(0, max_y_value * 1.05)  # Set y-axis range
        
        if i == 1:
            ax.legend(ncol=3)
            

    plt.tight_layout()
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()
    

def plot_best_prefetch_distance_per_locality(data, file_path_out: str, metric: str, is_max_better: bool, y_axis_label: str, hyper_threads: int, load_heavy: int, is_interleaved: bool, page_size: str):
    if 'hyper-threads' in data.columns:
        data = data.query(f'`hyper-threads` == {hyper_threads}')

    if 'page-size' in data.columns:
        data = data.query(f'`page-size` == "{page_size}"')

    if 'is-interleaved-prefetching' in data.columns:
        full_prefetch_data = data[
            (data['is-random'] == 1) &
            (data['prefetch-distance'] > 0) &
            (data['cache-lines'] == data['prefetched-cache-lines']) &
            (data['is-adjacent-pref'] == 0) &
            (data['is-interleaved-prefetching'] == int(is_interleaved))
        ]
    else:
        full_prefetch_data = data[
            (data['is-random'] == 1) &
            (data['prefetch-distance'] > 0) &
            (data['cache-lines'] == data['prefetched-cache-lines']) &
            (data['is-adjacent-pref'] == 0)
        ]
    
    if 'load-heavy' in data.columns:
        full_prefetch_data = full_prefetch_data.query(f'`load-heavy` == {load_heavy}')
    
    # Group by 'cache-lines' and 'prefetch-locality', and find the entry with the highest 'GB / second' for each group
    best_performance_indices = full_prefetch_data.groupby(['cache-lines', 'prefetch-locality'])[metric]
    if is_max_better:
        best_performance_indices = best_performance_indices.idxmax()
    else:
        best_performance_indices = best_performance_indices.idxmin()
    best_performance_by_locality = full_prefetch_data.loc[best_performance_indices]
    
    # Random and Sequential performance
    random_performance = data[(data['is-random'] == 1) & (data['prefetch-distance'] == 0)]
    sequential_performance = data[(data['is-random'] == 0)& (data['prefetch-locality'] == -1)]
    
    # Create a bar plot with adjusted text labels and renamed localities
    plt.figure(figsize=(14, 8))
    
    unique_cache_lines = full_prefetch_data['cache-lines'].unique()
    unique_prefetch_localities = full_prefetch_data['prefetch-locality'].unique()
    
    # Width of a bar
    width = 0.2
    
    # Iterate over renamed prefetch-locality values to create stacked bars
    for i, locality in enumerate(unique_prefetch_localities):
        # Filter data for the current prefetch-locality
        data_locality = best_performance_by_locality[best_performance_by_locality['prefetch-locality'] == locality]
        
        # Align bars for different localities
        x = np.arange(len(unique_cache_lines)) - (width * len(unique_prefetch_localities)) / 2 + i * width + width / 2
        
        # Create bar for the current prefetch-locality
        bars = plt.bar(x, data_locality[metric], width, label=f'{prefetch_locality_names[locality]}', color=color_mapping[prefetch_locality_names[locality]])
        
        # Add prefetch-distance as text above the bars
        for bar, prefetch_dist in zip(bars, data_locality['prefetch-distance']):
            yval = bar.get_height()
            plt.text(bar.get_x() + bar.get_width()/2, yval + 0.01, f'{prefetch_dist}', ha='center', va='bottom', rotation=45)
            
    # Plot horizontal lines for random execution throughput
    random_length = len(unique_prefetch_localities) / 2 * width
    for cache_line_index in range(0,len(unique_cache_lines)):
        cache_line = unique_cache_lines[cache_line_index]
        baseline_x = [cache_line_index-random_length,cache_line_index+random_length]
        
        # Random
        random_throughput = random_performance[random_performance['cache-lines'] == cache_line][metric].values[0]
        plt.plot(
            baseline_x, 
            [random_throughput,random_throughput], 
            label='Random' if cache_line_index == 0 else '', 
            color=color_mapping['Random'],
            linewidth=3
        )
        
        # Sequential
        sequential_throughput = sequential_performance[sequential_performance['cache-lines'] == cache_line][metric].values[0]
        plt.plot(
            baseline_x, 
            [sequential_throughput,sequential_throughput], 
            label='Sequential' if cache_line_index == 0 else '', 
            color=color_mapping['Sequential'],
            linewidth=3
        )

    # Add labels and title
    plt.xlabel('Block Size in Cache Lines')
    plt.ylabel(y_axis_label)
    #plt.title('Highest Throughput by Cache-Lines and Prefetch-Locality')
    plt.xticks(ticks=np.arange(len(unique_cache_lines)), labels=unique_cache_lines)
    plt.legend()

    # Save plot
    plt.tight_layout()
    plt.savefig(file_path_out)
    plt.close()
    print(f'Created {file_path_out}')
    plt.close()
    

def plot_best_partial_prefetch(data, file_path_out: str, metric: str, is_max_better: bool, y_axis_label: str, hyper_threads: int, load_heavy: int, is_interleaved: bool, page_size: str):
    if 'hyper-threads' in data.columns:
        data = data.query(f'`hyper-threads` == {hyper_threads}')

    if 'page-size' in data.columns:
        data = data.query(f'`page-size` == "{page_size}"')

    prefetch_data = data[(data['is-random'] == 1) & (data['prefetch-distance'] > 0) & (data['is-adjacent-pref'] == 0)]
   
    if 'is-interleaved-prefetching' in data.columns:
        prefetch_data = prefetch_data[prefetch_data['is-interleaved-prefetching'] == int(is_interleaved)]
    
    if 'load-heavy' in data.columns:
        prefetch_data = prefetch_data.query(f'`load-heavy` == {load_heavy}')
    
    prefetch_data = prefetch_data.groupby(['cache-lines', 'prefetched-cache-lines', 'prefetch-locality'])[metric]
    if is_max_better:
        prefetch_data = prefetch_data.max().reset_index()
    else:
        prefetch_data = prefetch_data.min().reset_index()
    
    max_y_value = data[metric].max()
    
    unique_cache_lines = prefetch_data['cache-lines'].unique()
    unique_prefetch_localities = prefetch_data['prefetch-locality'].unique()
    
    plt.figure(figsize=(15, 12))
    
    # Creating the grouped bar plot manually
    bar_width = 0.15  # Width of the bars

    num_rows = int(math.ceil(len(unique_cache_lines) / 3))

    for i, cache_line_size in enumerate(unique_cache_lines, 1):
        ax = plt.subplot(num_rows, 3, i)
        
        cache_line_data = prefetch_data[prefetch_data['cache-lines'] == cache_line_size]
        indices = np.arange(len(cache_line_data['prefetched-cache-lines'].unique()))  # x locations for the groups
        
        for j, locality in enumerate(unique_prefetch_localities):
            locality_data = cache_line_data[cache_line_data['prefetch-locality'] == locality]
            ax.bar(
                indices + j * bar_width, 
                locality_data[metric], 
                bar_width, 
                label=prefetch_locality_names[locality],
                color=color_mapping[prefetch_locality_names[locality]]
            )
            
        # Add horizontal (base-) line for random and sequential
        sequential_data = data[(data['is-random'] == 0) & (data['cache-lines'] == cache_line_size) & (data['prefetch-locality'] == -1)].reset_index()
        if len(sequential_data) > 0:
            ax.axhline(sequential_data.at[0, metric], color=color_mapping['Sequential'], linestyle='-', label='Sequential')
            
        random_data = data[(data['is-random'] == 1) & (data['prefetch-distance'] == 0) & (data['cache-lines'] == cache_line_size)].reset_index()
        if len(random_data) > 0:
            ax.axhline(random_data.at[0, metric], color=color_mapping['Random'], linestyle='-', label='Random')
            
        ax.set_title(f'Cache-Line Size: {cache_line_size}')
        ax.set_ylabel(y_axis_label)
        ax.set_ylim(0, max_y_value * 1.05)  # Set y-axis range
        ax.set_xlabel('Prefetched Cache Lines')
        ax.set_xticks(indices + bar_width / len(unique_prefetch_localities))
        ax.set_xticklabels(locality_data['prefetched-cache-lines'])
        if i == 1:
            ax.legend(ncol=3)
        
    plt.tight_layout()
    plt.subplots_adjust(top=0.9)
    
    # Save plot
    plt.tight_layout()
    plt.savefig(file_path_out)
    plt.close()
    print(f'Created {file_path_out}')
    plt.close()
    
    
def plot_adjacent_prefetch(data, file_path_out: str, metric: str, is_max_better: bool, y_axis_label: str, hyper_threads: int, load_heavy: int, is_interleaved: bool, page_size: str):
    if 'hyper-threads' in data.columns:
        data = data.query(f'`hyper-threads` == {hyper_threads}')

    if 'page-size' in data.columns:
        data = data.query(f'`page-size` == "{page_size}"')
    
    # Filter the data for prefetching benchmarks and divide into normal and adjacent prefetching
    prefetching_data = data[(data['is-random'] == 1) & (data['prefetch-distance'] > 0)]
    if 'is-interleaved-prefetching' in data.columns:
        prefetching_data = prefetching_data[(prefetching_data['is-interleaved-prefetching'] == int(is_interleaved))]
    if 'load-heavy' in data.columns:
        prefetching_data = prefetching_data.query(f'`load-heavy` == {load_heavy}')

    normal_prefetching_data = prefetching_data[(prefetching_data['is-adjacent-pref'] == 0) & (prefetching_data['cache-lines'] == prefetching_data['prefetched-cache-lines'])]
    adjacent_prefetching_data = prefetching_data[(prefetching_data['is-adjacent-pref'] == 1)]
    
    # Random and Sequential performance
    random_performance = data[(data['is-random'] == 1) & (data['prefetch-distance'] == 0)]
    sequential_performance = data[(data['is-random'] == 0) & (data['prefetch-locality'] == -1)]

    # Group by cache-lines and find the row with the max throughput for each group
    best_normal_prefetching_indices = normal_prefetching_data.groupby('cache-lines')[metric]
    if is_max_better:
        best_normal_prefetching_indices = best_normal_prefetching_indices.idxmax()
    else:
        best_normal_prefetching_indices = best_normal_prefetching_indices.idxmin()
    best_normal_prefetching = normal_prefetching_data.loc[best_normal_prefetching_indices][['cache-lines', metric, 'prefetch-locality', 'prefetch-distance']]

    best_adjacent_prefetching_indices = adjacent_prefetching_data.groupby('cache-lines')[metric]
    if is_max_better:
        best_adjacent_prefetching_indices = best_adjacent_prefetching_indices.idxmax()
    else:
        best_adjacent_prefetching_indices = best_adjacent_prefetching_indices.idxmin()

    best_adjacent_prefetching = adjacent_prefetching_data.loc[best_adjacent_prefetching_indices][['cache-lines', metric, 'prefetch-locality', 'prefetch-distance']]

    comparison_table = pd.merge(best_normal_prefetching, best_adjacent_prefetching, on='cache-lines', how='outer').query('`cache-lines` > 1')

    # Plot bars
    fig, ax = plt.subplots(figsize=(14, 8))
    bar_width = 0.35
    indices = range(len(comparison_table))
    normal_bars = ax.bar(indices, comparison_table[f'{metric}_x'], bar_width, label='Full Prefetching', color=color_mapping['Full'])
    adjacent_bars = ax.bar([i + bar_width for i in indices], comparison_table[f'{metric}_y'], bar_width, label='Adjacent Prefetching', color=color_mapping['Adjacent'])
    
    # Plot horizontal lines for random execution throughput
    random_length = bar_width
    unique_cache_lines = comparison_table['cache-lines'].unique()
    for cache_line_index in range(0,len(unique_cache_lines)):
        cache_line = unique_cache_lines[cache_line_index]
        baseline_x = [cache_line_index-random_length+bar_width/2,cache_line_index+random_length+bar_width/2]
        
        # Random
        random_throughput = random_performance[random_performance['cache-lines'] == cache_line][metric].values[0]
        ax.plot(
            baseline_x, 
            [random_throughput,random_throughput], 
            label='Random' if cache_line_index == 0 else '', 
            color=color_mapping['Random'],
            linewidth=3
        )
        
        # Sequential
        sequential_throughput = sequential_performance[sequential_performance['cache-lines'] == cache_line][metric].values[0]
        ax.plot(
            baseline_x, 
            [sequential_throughput,sequential_throughput], 
            label='Sequential' if cache_line_index == 0 else '', 
            color=color_mapping['Sequential'],
            linewidth=3
        )

    ax.set_xlabel('Cache-Lines', fontsize=14)
    ax.set_ylabel(y_axis_label, fontsize=14)
    ax.set_xticks([i + bar_width / 2 for i in indices])
    ax.set_xticklabels(comparison_table['cache-lines'])
    ax.legend()
    
    # Save plot
    plt.tight_layout()
    plt.savefig(file_path_out)
    plt.close()
    print(f'Created {file_path_out}')
    plt.close()
    
def plot_partial_cycles(data, file_path_out: str, hyper_threads: int, load_heavy: int, is_interleaved: bool, page_size: str):
    if 'hyper-threads' in data.columns:
        data = data.query(f'`hyper-threads` == {hyper_threads}')

    if 'page-size' in data.columns:
        data = data.query(f'`page-size` == "{page_size}"')

    # Create cycle columns
    data['cycles-no-stall'] = data['cycles'] - data['CYCLE_ACTIVITY.STALLS_TOTAL']
    data['cycles-mem-stall'] = data['CYCLE_ACTIVITY.STALLS_MEM_ANY']
    data['cycles-other-stall'] = data['CYCLE_ACTIVITY.STALLS_TOTAL'] - data['CYCLE_ACTIVITY.STALLS_MEM_ANY']

    if 'is-interleaved-prefetching' in data.columns:
        # Filter only (relevant) prefetches
        prefetch_data = data[
            (data['is-random'] == 1) &
            (data['prefetch-distance'] > 0) &
            (data['is-adjacent-pref'] == 0) &
            (data['hyper-threads'] == 1) &
            (data['is-interleaved-prefetching'] == int(is_interleaved))
        ]
    else:
        # Filter only (relevant) prefetches
        prefetch_data = data[
            (data['is-random'] == 1) &
            (data['prefetch-distance'] > 0) &
            (data['is-adjacent-pref'] == 0) &
            (data['hyper-threads'] == 1)
        ]
        
    if 'load-heavy' in data.columns:
        prefetch_data = prefetch_data.query(f'`load-heavy` == {load_heavy}')

    # Filter only the best distances
    prefetch_data_indices = prefetch_data.groupby(['cache-lines', 'prefetch-locality', 'prefetched-cache-lines'])['cycles'].idxmin()
    prefetch_data = prefetch_data.loc[prefetch_data_indices]
    #max_y_value = prefetch_data['cycles'].max()
    max_y_value = 100 if np.isnan(prefetch_data['cycles'].max()) or np.isinf(prefetch_data['cycles'].max()) else prefetch_data['cycles'].max()

    unique_cache_lines = data['cache-lines'].unique()
    unique_prefetch_localities = data['prefetch-locality'].unique()

    plt.figure(figsize=(24, 12))
    
    # Creating the grouped bar plot manually
    bar_width = 0.2  # Width of the bars

    num_rows = int(math.ceil(len(unique_cache_lines) / 3))

    legend_bars = {}

    for i, cache_line_size in enumerate(unique_cache_lines, 1):
        ax = plt.subplot(num_rows, 3, i)
        
        cache_line_data = prefetch_data[prefetch_data['cache-lines'] == cache_line_size]
        cache_line_group = cache_line_data.pivot(index='prefetched-cache-lines', columns='prefetch-locality', values=['cycles-no-stall','cycles-mem-stall','cycles-other-stall'])
        
        x = np.arange(len(cache_line_data['prefetched-cache-lines'].unique()))
        
        bottoms = {}
        bottom = None
        
        labels_x = {}
        
        for j, (column, column_df) in enumerate(cache_line_group.items()):
            label = prefetch_locality_names[column[1]]
            
            color = color_mapping[label]
            fill_pattern = fill_mapping[column[0]]
            
            if label in bottoms:
                bottom = bottoms[label]
                bottoms[label] = bottom + column_df
            else:
                bottom = None
                bottoms[label] = column_df
            
            if label in labels_x:
                x_pos = labels_x[label]
            else:
                x_pos = j
                labels_x[label] = j
            
            bar = ax.bar(.1 + x + (x_pos - 4 / 2) * bar_width, column_df, bar_width, bottom=bottom, label=label, color=color, hatch=fill_pattern, edgecolor='black')
            if label not in legend_bars:
                legend_bars[label] = bar
            
        sequential_data = data[(data['is-random'] == 0) & (data['cache-lines'] == cache_line_size) & (data['prefetch-locality'] == -1)].reset_index()
        if len(sequential_data) > 0:
            ax.axhline(sequential_data.at[0, 'cycles'], color=color_mapping['Sequential'], linestyle='-', label='Sequential')
            
        ax.set_title(f'block = {cache_line_size} cache lines')
        
        # Grid
        ax.grid(color='gray', linestyle='-', linewidth=.5, axis='y')
        ax.set_axisbelow(True)
        
        # No border on top
        #ax.spines.right.set_visible(False)
        #ax.spines.top.set_visible(False)
        
        
        # Labels
        ax.set_xlabel('prefetched lines', fontsize=13)
        ax.set_ylabel('cycles / cache line', fontsize=13)
        ax.set_ylim(0, max_y_value)
        
        # Ticks
        ax.tick_params(labelsize=15, axis='y')
        ax.set_xticks(x + bar_width / len(unique_prefetch_localities))
        ax.set_xticklabels(cache_line_data['prefetched-cache-lines'].unique())
        
        # Legend
        if i == 1:
            legend_handles = [handle for key, handle in legend_bars.items()]
            legend_handles.append(patches.Patch(edgecolor='black', fill=False, label='mem stalls', hatch=fill_mapping['cycles-mem-stall']))
            legend_handles.append(patches.Patch(edgecolor='black', fill=False, label='other stalls', hatch=fill_mapping['cycles-other-stall']))
            plt.legend(loc='upper center', handles=legend_handles, ncol=3, fontsize=11) #
        
    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.85)
    plt.tight_layout()
    
    # Save plot
    plt.savefig(file_path_out)
    plt.close()
    print(f'Created {file_path_out}')
    plt.close()
    
## Plot
parser = argparse.ArgumentParser()
parser.add_argument('--in', dest='in_file', help='File to plot', required=True)
parser.add_argument('--out', dest='out_dir', help='Folder to save plots (default will be the folder of the in-file).', default=None)

args = parser.parse_args()
out = args.out_dir if args.out_dir else os.path.dirname(args.in_file)

data = pd.read_csv(args.in_file)


ht = [1] if 'hyper-threads' not in data.columns else data['hyper-threads'].unique()
ht.sort()
page_sizes = ['4K'] if 'page-size' not in data.columns else ['4K', '2M', '1G']
interleaved = [False] if 'is-interleaved-prefetching' not in data.columns else [False, True]
lh = [0] if 'load-heavy' not in data.columns else [0, 1]


for hyper_threads in ht:
    hyper_threads_suffix = '-hyperthreads' if hyper_threads == 2 else ''
    for load_heavy in lh:
        if hyper_threads == 1 and load_heavy == 1:
            print("Combination not possible")
            break
        load_heavy_suffix = hyper_threads_suffix
        load_heavy_suffix += '-load-heavy' if load_heavy == 1 else ''
        for page_size in page_sizes:
            for is_interleaved in interleaved:
                suffix = load_heavy_suffix
                if is_interleaved:
                    suffix += '-interleaved'
                suffix += '-page-size-' + str(page_size)

                plot_prefetch_distances(data, os.path.join(out, f'throughput-prefetch-distance{suffix}.pdf'), 'GB / second', 'Throughput (GB/s)', hyper_threads, load_heavy, is_interleaved, page_size)

                if not is_interleaved:
                    plot_best_prefetch_distance_per_locality(data, os.path.join(out, f'throughput-per-block-and-locality{suffix}.pdf'), 'GB / second', True, 'Throughput (GB/s)', hyper_threads, load_heavy, is_interleaved, page_size)

                plot_best_partial_prefetch(data, os.path.join(out, f'throughput-partial-prefetches{suffix}.pdf'), 'GB / second', True, 'Throughput (GB/s)', hyper_threads, load_heavy, is_interleaved, page_size)
                if 'CYCLE_ACTIVITY.STALLS_TOTAL' in data.columns:
                    plot_best_partial_prefetch(data, os.path.join(out, f'stalls-total-partial-prefetches{suffix}.pdf'), 'CYCLE_ACTIVITY.STALLS_TOTAL', False, 'Stalled Cycles / Cache Line', hyper_threads, load_heavy, is_interleaved, page_size)

                #if not is_interleaved:
                #    plot_adjacent_prefetch(data, os.path.join(out, f'throughput-adjacent-prefetching{suffix}.pdf'), 'GB / second', True, 'Throughput (GB/s)', hyper_threads, load_heavy, is_interleaved, page_size)

                if hyper_threads == 1 and 'cycles' in data.columns and 'CYCLE_ACTIVITY.STALLS_TOTAL' in data.columns and 'CYCLE_ACTIVITY.STALLS_MEM_ANY' in data.columns:
                    plot_partial_cycles(data, os.path.join(out, f'cycle-breakdown{suffix}.pdf'), hyper_threads, load_heavy, is_interleaved, page_size)
