#!/usr/bin/python3

import argparse
import json
import os
import subprocess
import pandas
import signal
import sys


class Result:
    _results = {}
    _counter_names = []

    def __init__(self, counter_names: list):
        self._counter_names = counter_names

        self._results['cache-lines'] = []
        self._results['is-random'] = []
        self._results['prefetch-locality'] = []
        self._results['prefetch-distance'] = []
        self._results['prefetched-cache-lines'] = []
        self._results['is-l1-pref'] = []
        self._results['is-l2-pref'] = []
        self._results['ns'] = []
        self._results['GB / second'] = []
        for counter_name in self._counter_names:
            self._results[counter_name] = []

    def add(self, is_random: bool, cache_lines: int, prefetch_locality: int, prefetch_distance: int, prefetched_cache_lines: int, is_l1_pref: bool, is_l2_pref: bool, results):
        self._results['cache-lines'].append(cache_lines)
        self._results['is-random'].append(int(is_random))
        self._results['prefetch-locality'].append(prefetch_locality)
        self._results['prefetch-distance'].append(prefetch_distance)
        self._results['prefetched-cache-lines'].append(prefetched_cache_lines)
        self._results['is-l1-pref'].append(int(is_l1_pref))
        self._results['is-l2-pref'].append(int(is_l2_pref))
        self._results['ns'].append(results['ns'])
        self._results['GB / second'].append(results['GB / second'])
        for counter_name in self._counter_names:
            self._results[counter_name].append(results[counter_name])

    def to_data_frame(self):
        return pandas.DataFrame(self._results)

def exec_with_params(params, temp_file='tmp.json'):
    if os.path.exists(temp_file):
        os.remove(temp_file)
    subprocess.call(f'./bin/swpf_benchmark {params}', shell=True)
    with open(temp_file, 'r') as result_json_file:
        result = json.loads(result_json_file.read())
        return result


parser = argparse.ArgumentParser()
parser.add_argument("--profile", type=str, help="Perf profile", default=None)
parser.add_argument("--gb", type=int, help="Size of the global array to read.", default=2)
parser.add_argument("--out", dest="out_file", type=str, help="Name of the out file.", default="out-hw-sw-pref.csv")
parser.add_argument("--intel", help="Indicate an Intel machine; used to disable prefetchers.", action='store_true')
parser.add_argument("--zen4", help="Indicate an AMD Zen4 machine; used to disable prefetchers.", action='store_true')
args = parser.parse_args()

if not args.intel and not args.zen4:
    print("Unknown machine; use --intel or --zen4")
    exit(1)


def set_prefetcher(is_l1, is_l2, is_intel, is_zen4):
    if is_intel:
        core_prefetcher_msr = 12 if not is_l1 else 0
        if not is_l2:
            core_prefetcher_msr += 3
        subprocess.call(f'sudo wrmsr -a 0x1A4 {core_prefetcher_msr}', shell=True)
    elif is_zen4:
        core_prefetcher_msr = 0 if is_l1 else 7
        if not is_l2:
            core_prefetcher_msr += 40

        subprocess.call(f'sudo wrmsr -a 0xC0000108 {core_prefetcher_msr}', shell=True)

## Set benchmark environment.
#subprocess.call('sudo sh -c "echo 0 > /proc/sys/kernel/numa_balancing"', shell=True)
#subprocess.call('sudo cpupower frequency-set -g performance', shell=True)
#subprocess.call('sudo sysctl -w kernel.perf_event_paranoid=-1', shell=True)

arguments_template = [f'--gb {args.gb}', "--out tmp.json", '--result-per-cacheline']

counter = []
if args.profile and os.path.exists(args.profile):
    with open(args.profile, 'r') as profile_file:
        data = json.loads(profile_file.read())
        if "counter" in data:
            counter = data["counter"]
    arguments_template.append(f'--perf-profile {args.profile}')

result = Result(counter)

# Prefetch localities
prefetch_localities = [0, 1, 2, 3]

for is_l1_pref in [False, True]:
    for is_l2_pref in [False, True]:

        # Disable Prefetcher
        set_prefetcher(is_l1_pref, is_l2_pref, args.intel, args.zen4)

        for block_size in [16]:
            for prefetched_cache_lines in ([1] + list(range(2, block_size + 1, 2))):
                for prefetch_locality in prefetch_localities:
                    arguments = arguments_template.copy()
                    arguments.append('--random')
                    arguments.append(f'--cache-lines {block_size}')
                    arguments.append(f'--prefetch-lines {prefetched_cache_lines}')
                    arguments.append(f'--prefetch-distance 1')
                    arguments.append(f'--prefetch-locality {prefetch_locality}')
                    arguments.append(f'--cores "0"')
                    arguments.append('--interleaved')

                    prefetched_execution = exec_with_params(' '.join(arguments))
                    result.add(True, block_size, prefetch_locality, 1, prefetched_cache_lines, is_l1_pref, is_l2_pref, prefetched_execution)

# Disable Prefetcher
set_prefetcher(True, True, args.intel, args.zen4)

data = result.to_data_frame()
data.to_csv(args.out_file)
print(f'Wrote {len(data)} experiments to {args.out_file}')
