#!/usr/bin/python3

import argparse
import json
import os
import subprocess
import pandas
import signal
import sys


class Result:
    _results = {}
    _counter_names = []

    def __init__(self, counter_names: list):
        self._counter_names = counter_names

        self._results['cache-lines'] = []
        self._results['hyper-threads'] = []
        self._results['load-heavy'] = []
        self._results['page-size'] = []
        self._results['is-random'] = []
        self._results['prefetch-locality'] = []
        self._results['prefetch-distance'] = []
        self._results['prefetched-cache-lines'] = []
        self._results['is-adjacent-pref'] = []
        self._results['is-double-prefetching'] = []
        self._results['double-prefetch-distance'] = []
        self._results['is-interleaved-prefetching'] = []
        self._results['ns'] = []
        self._results['GB / second'] = []
        for counter_name in self._counter_names:
            self._results[counter_name] = []

    def add(self, is_random: bool, cache_lines: int, hyper_threads: int, load_heavy: bool, page_size: str, prefetch_locality: int, prefetch_distance: int, prefetched_cache_lines: int, is_adjacent_pref: bool, is_double_prefetch: int, double_prefetch_distance: int, is_interleaved_prefetch: bool, results):
        self._results['cache-lines'].append(cache_lines)
        self._results['hyper-threads'].append(hyper_threads)
        self._results['load-heavy'].append(load_heavy)
        self._results['page-size'].append(page_size)
        self._results['is-random'].append(int(is_random))
        self._results['prefetch-locality'].append(prefetch_locality)
        self._results['prefetch-distance'].append(prefetch_distance)
        self._results['prefetched-cache-lines'].append(prefetched_cache_lines)
        self._results['is-adjacent-pref'].append(int(is_adjacent_pref))
        self._results['is-double-prefetching'].append(int(is_double_prefetch))
        self._results['double-prefetch-distance'].append(int(double_prefetch_distance))
        self._results['is-interleaved-prefetching'].append(int(is_interleaved_prefetch))
        self._results['ns'].append(results['ns'])
        self._results['GB / second'].append(results['GB / second'])
        for counter_name in self._counter_names:
            self._results[counter_name].append(results[counter_name])

    def to_data_frame(self):
        return pandas.DataFrame(self._results)

def exec_with_params(params, temp_file='tmp.json'):
    if os.path.exists(temp_file):
        os.remove(temp_file)
    subprocess.call(f'./bin/swpf_benchmark {params}', shell=True)
    try:
        with open(temp_file, 'r') as result_json_file:
            result = json.loads(result_json_file.read())
            return result
    except Exception as e:
        print("Program with parameters " + str(params) + " crashed. Print all results so far...")
        try:
            data = result.to_data_frame()
            data.to_csv(args.out_file)
            print(f'Wrote {len(data)} experiments to {args.out_file}')
        except Exception as f:
            print("Error, could not write data...", f)
        try:
            with open('error_log.txt', 'a') as file:
                file.write("Program with params: " + str(params) + " crashed\n")
        except Exception as g:
            print("Error could not create an error log...", g)
        

def thread_siblings(physical_cpu_id):
    with open(f'/sys/devices/system/cpu/cpu{physical_cpu_id}/topology/thread_siblings_list', 'r') as file:
        content = file.read().strip()
    parts = content.split(',')
    thread_siblings = []
    for part in parts:
        if '-' in part:
            start, end = map(int, part.split('-'))
            thread_siblings.extend(range(start, end + 1))
        else:
            thread_siblings.append(int(part))
    return thread_siblings

# Write all the results so far
def signal_handler(signal, frame):
    print("Print all results so far...")
    data = result.to_data_frame()
    data.to_csv(args.out_file)
    print(f'Wrote {len(data)} experiments to {args.out_file}')
    sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)


parser = argparse.ArgumentParser()
parser.add_argument("--profile", type=str, help="Perf profile", default=None)
parser.add_argument("--gb", type=int, help="Size of the global array to read.", default=2)
parser.add_argument("--out", dest="out_file", type=str, help="Name of the out file.", default="out.csv")
parser.add_argument("--t2", help="If set, will execute benchmarks with T2 instruction.", action='store_true')
parser.add_argument("--no-prefetching", help="If set, will skip prefetching runs.", action='store_true')
parser.add_argument("--interleaved-prefetching", help="If set, will execute the interleaved prefetching runs.", action='store_true')
parser.add_argument("--huge-pages", help="If set, will execute additional benchmarks with huge pages (2M, 1G)", action='store_true')
parser.add_argument("--adjacent-prefetching", help="If set, will execute also adjacent prefetching benchmark.", action='store_true')
parser.add_argument("--trend-only", help="If set, will execute only benchmarks needed for the trend-plot.", action='store_true')
args = parser.parse_args()

## Set benchmark environment.
#subprocess.call('sudo sh -c "echo 0 > /proc/sys/kernel/numa_balancing"', shell=True)
#subprocess.call('sudo cpupower frequency-set -g performance', shell=True)
#subprocess.call('sudo sysctl -w kernel.perf_event_paranoid=-1', shell=True)

arguments_template = [f'--gb {args.gb}', "--out tmp.json", '--result-per-cacheline']

counter = []
if args.profile and os.path.exists(args.profile):
    with open(args.profile, 'r') as profile_file:
        data = json.loads(profile_file.read())
        if "counter" in data:
            counter = data["counter"]
    arguments_template.append(f'--perf-profile {args.profile}')

result = Result(counter)

# Prefetch localities; include T2 if set (opt in)
prefetch_localities = [0, 1, 3] if not args.t2 else [0, 1, 2, 3]

# Maps from block size to tested prefetch distances.
block_sizes_and_prefetch_distances = {
    1: [1,2,4,8,16,24,32],
    2: [1,2,4,8,16,24],
    4: [1,2,4,8,16],
    8: [1,2,4,8],
    16: [1,2,4],
    32: [1,2]
}

# Test only a single cache line per block for the trend (opt in)
block_sizes = [1, 2, 4, 8, 16, 32] if not args.trend_only else [1]

# If desired, test also huge pages of 2M and 1G (opt in)
memory_page_configurations = ['4K'] if not args.huge_pages else ['4K', '2M', '1G']

## Only on a single core AND on all hyper threads
core_configurations = [ [0] ]
core_configurations.reverse()

for cores in core_configurations:
    core_list = ' '.join([str(c) for c in cores])
    for memory_page_size in memory_page_configurations:
        print("Core(s) used: " + str(cores) + ", Pagesize used: " + str(memory_page_size))
        ## Sequential
        for cache_lines in block_sizes:
            arguments = arguments_template.copy()
            arguments.append(f'--cache-lines {cache_lines}')
            arguments.append(f'--cores "{core_list}"')
            arguments.append(f'--page-size {memory_page_size}')

            sequential_execution = exec_with_params(' '.join(arguments))
            result.add(False, cache_lines, len(cores), 0, memory_page_size, -1, 0, 0, False, False, 0, False, sequential_execution)

            ### Sequential + SWPF for measuring instructions for SWPF.
            if not args.trend_only:
                prefetch_distance = 1
                for prefetch_locality in prefetch_localities:
                    arguments = arguments_template.copy()
                    arguments.append(f'--cache-lines {cache_lines}')
                    arguments.append(f'--cores "{core_list}"')
                    arguments.append(f'--prefetch-distance {prefetch_distance}')
                    arguments.append(f'--prefetch-locality {prefetch_locality}')
                    arguments.append(f'--page-size {memory_page_size}')

                    sequential_execution_with_prefetching = exec_with_params(' '.join(arguments))
                    result.add(False, cache_lines, len(cores), 0, memory_page_size, prefetch_locality, prefetch_distance, cache_lines, False, False, 0, False, sequential_execution_with_prefetching)

        ## Random
        for cache_lines in block_sizes:
            arguments = arguments_template.copy()
            arguments.append('--random')
            arguments.append(f'--cache-lines {cache_lines}')
            arguments.append(f'--cores "{core_list}"')
            arguments.append(f'--page-size {memory_page_size}')

            random_execution = exec_with_params(' '.join(arguments))
            result.add(True, cache_lines, len(cores), 0, memory_page_size, -1, 0, 0, False, False, 0, False, random_execution)


        ## SWPF
        if not args.no_prefetching:
            # Sequential Prefetching
            for cache_lines in block_sizes:
                print("Core(s) used: " + str(cores) + ", Pagesize used: " + str(memory_page_size) + ", Cacheline: " + str(cache_lines))
                for prefetched_cache_lines in ([1] + list(range(2,max(block_sizes) + 1, 2))):
                    print("Core(s) used: " + str(cores) + ", Pagesize used: " + str(memory_page_size) + ", CLs: " + str(cache_lines) + ", pf CLs: " + str(prefetched_cache_lines))
                    if prefetched_cache_lines <= cache_lines:
                        for prefetch_distance in block_sizes_and_prefetch_distances[cache_lines]:
                            print("Core(s) used: " + str(cores) + ", Pagesize used: " + str(memory_page_size) + ", CLs: " + str(cache_lines) + ", pf CLs: " + str(prefetched_cache_lines) + ", pf distance: " + str(prefetch_distance))
                            for prefetch_locality in prefetch_localities:
                                print("Core(s) used: " + str(cores) + ", Pagesize used: " + str(memory_page_size) + ", CLs: " + str(cache_lines) + ", pf CLs: " + str(prefetched_cache_lines) + ", pf distance: " + str(prefetch_distance) + ", pf loc: " + str(prefetch_locality))
                                arguments = arguments_template.copy()
                                arguments.append('--random')
                                arguments.append(f'--cache-lines {cache_lines}')
                                arguments.append(f'--prefetch-lines {prefetched_cache_lines}')
                                arguments.append(f'--prefetch-distance {prefetch_distance}')
                                arguments.append(f'--prefetch-locality {prefetch_locality}')
                                arguments.append(f'--cores "{core_list}"')
                                arguments.append(f'--page-size {memory_page_size}')

                                prefetched_execution = exec_with_params(' '.join(arguments))
                                result.add(True, cache_lines, len(cores), 0, memory_page_size, prefetch_locality, prefetch_distance, prefetched_cache_lines, False, False, 0, False, prefetched_execution)

                                if args.interleaved_prefetching:
                                    arguments.append('--interleaved')
                                    prefetched_execution = exec_with_params(' '.join(arguments))
                                    result.add(True, cache_lines, len(cores), 0, memory_page_size, prefetch_locality, prefetch_distance, prefetched_cache_lines, False, False, 0, True, prefetched_execution)



        ## Adjacent Prefetching (only with --adjacent-prefetching)
        if args.adjacent_prefetching and not args.trend_only:
            for cache_lines in block_sizes:
                for prefetch_distance in block_sizes_and_prefetch_distances[cache_lines]:
                    for prefetch_locality in prefetch_localities:
                        arguments = arguments_template.copy()
                        arguments.append('--random')
                        arguments.append(f'--cache-lines {cache_lines}')
                        arguments.append(f'--prefetch-lines {prefetched_cache_lines}')
                        arguments.append(f'--prefetch-distance {prefetch_distance}')
                        arguments.append(f'--prefetch-locality {prefetch_locality}')
                        arguments.append('--adjacent')
                        arguments.append(f'--cores "{core_list}"')
                        arguments.append(f'--page-size {memory_page_size}')

                        prefetched_execution = exec_with_params(' '.join(arguments))
                        result.add(True, cache_lines, len(cores), 0, memory_page_size, prefetch_locality, prefetch_distance, prefetched_cache_lines, True, 0, 0, prefetched_execution)



data = result.to_data_frame()
data.to_csv(args.out_file)
print(f'Wrote {len(data)} experiments to {args.out_file}')
