import pandas as pd
import matplotlib.pyplot as plt
import argparse
import os
import matplotlib.patches as patches
import numpy as np
import matplotlib.ticker as mticker


#############################################################
#   Overall Settings                                        #
#                                                           #
#############################################################
plt.rcParams.update({
    "font.family": "serif",
    "text.usetex": True,
    'text.latex.preamble': "\\usepackage{libertine}\\usepackage[libertine]{newtxmath}",
})

#############################################################
#   Global Definitions                                      #
#                                                           #
#############################################################
colors = {
    'Intel': '#004477', #0068b5
    'AMD': '#00a874',
    'ARM': '#33a3c8', #'#008fbe'
    'ARM Server': '#33a3c8', 

    'prefetchNTA': '#e76f51',
    'prefetchT2': '#168aad',
    'prefetchT1': '#168aad',
    'prefetchT0': '#57cc99',

    'samples': '#555555',

    'random': '#555555',
    'sw-pref. 8 lines': '#57cc99',
    'sw-pref. 10 lines': '#168aad',
    'sw-pref. 12 lines': '#e76f51',

    'prefetched-lines': '#444444',

    '256 B node size': '#e76f51',
    '256 B node size (SMT)': '#f4a261',
    '1 kB node size': '#1e6091',
    '1 kB node size (SMT)': '#a7d9f1',
    '4 kB node size': '#57cc99',
    '4 kB node size (SMT)': '#c7f9cc',
}

markers = {
    'Intel': 'v',
    'AMD': 'o',
    'ARM': 's',
    'ARM Server': 'd', 

    'random': 'x',
    'sw-pref. 8 lines': 'X',
    'sw-pref. 10 lines': 'd',
    'sw-pref. 12 lines': '*',

    'prefetchNTA': '*',
    'prefetchT1': 'd',
    'prefetchT0': 'X',
}

bar_fills = {
    'overhead-compute-cycles': '',
    'stalled-cycles-mem': '//',
    'stalled-cycles-other': '...',
    'smt': '///'
}

prefetch_locality_names = {
    0: 'prefetchNTA',
    1: 'prefetchT1', #T2
    2: 'prefetchT1',
    3: 'prefetchT0'
}

column_names = {
    'overhead-compute-cycles': 'non-stalling cycles',
    'stalled-cycles-mem': 'memory stalled cycles',
    'stalled-cycles-other': 'other stalled cycles',
    'stalled-cycles': 'stalled cycles',
    'overhead-cycles': 'cycles',
    'overhead-ns': 'nanoseconds'
}


#############################################################
#   Trend                                                   #
#                                                           #
#############################################################
def get_trend_prefetching_improvement(file_path):
    data = pd.read_csv(file_path)
    random_condition = (data["is-random"] == 1) & (data["prefetch-distance"] == 0) & (data["cache-lines"] == 1)
    if "hyper-threads" in data.columns:
        random_condition = random_condition & (data["hyper-threads"] == 1)
    random_max = data.loc[random_condition, "GB / second"].max()
    
    prefetching_condition = (data["is-random"] == 1) & (data["prefetch-distance"] > 0) & (data["cache-lines"] == 1) & (data["prefetched-cache-lines"] == 1)
    if "hyper-threads" in data.columns:
        prefetching_condition = prefetching_condition & (data["hyper-threads"] == 1)
    prefetching_max = data.loc[prefetching_condition, "GB / second"].max()
    
    return 100.0 / random_max * prefetching_max / 100.0


def trend(vendors, out_dir):
    fig, ax_left = plt.subplots(figsize=(7,4))

    print(vendors)

    legend_handles = []

    # SWPF Improvement
    max_performance = 0
    ax_left.set_xlabel('processor release year', fontsize=16)
    ax_left.set_ylabel('speedup with SW prefetching\n(over pure random access)', fontsize=16)
    ax_left.yaxis.set_major_formatter(r'{x:1.0f}$\times$')
    ax_left.tick_params(axis='both', which='major', labelsize=16)

    for name, vendor in vendors.items():
        architectures = vendor[0]
        color = vendor[1]
        marker = vendor[2]

        years = [t[0] for t in architectures]
        names = [t[1] for t in architectures]
        performance = [t[2] for t in architectures]
        max_performance = max(max(performance), max_performance)

        handle, = ax_left.plot(years, performance, marker=marker, linestyle=':', color=color, linewidth=3, markersize=8, label=name)
        legend_handles.append(handle)  # Creates a line plot with markers


        xlabel_offset = [t[3] for t in architectures]
        ylabel_offset = [t[4] for t in architectures]
        for i, value in enumerate(performance):
            plt.text(years[i] + xlabel_offset[i], value + (max_performance * ylabel_offset[i]), names[i], ha='center', va='bottom', fontsize=13, color=colors[name])

    ax_left.set_ylim([1,max_performance * 1.1])
    ax_left.set_xlim(2010.5,2024)

    ax_left.grid(True, color='#aaaaaa', linestyle='--', axis='y', linewidth=.4)

    ax_left.legend(loc='upper left', handles=legend_handles, ncol=2, fontsize=15)
    ax_left.spines[['right', 'top']].set_visible(False)

    ax_left.yaxis.set_tick_params(labelsize=16)
    ax_left.xaxis.set_tick_params(labelsize=16)

    # Arrow
    ax_left.annotate('better', xy=(2023,2.25), xytext=(2023,1.5), ha='center', fontsize=14, arrowprops={'edgecolor': 'black', 'facecolor': 'black', 'arrowstyle': '-|>', 'lw': 3})

    #ax_left.legend()
    plt.tight_layout()
    out_file = os.path.join(out_dir, 'trend.pdf')
    plt.savefig(out_file)
    print(f'Created {out_file}')


#############################################################
#   Stalls Breakdown                                        #
#                                                           #
#############################################################
def stalls_breakdown(data, file_path_out: str, metrics: list, hyper_threads: int, cache_lines: int, is_interleaved: int):

    plot_data = data.query(f'`cache-lines` == {cache_lines}')
    if 'hyper-threads' in data.columns:
        plot_data = plot_data.query(f'`hyper-threads` == {hyper_threads}')

    if 'is-interleaved-prefetching' in data.columns:
        plot_data = plot_data.query(f'`is-interleaved-prefetching` == {is_interleaved}')

    ## Sequential run to compare prefetched runs with
    sequential = plot_data.query('`is-random` == 0 & `prefetch-locality` == -1').iloc[0]

    ## Calculate overhead columns based in the available metrics (STALLS + MEM_STALLS, STALLS, CYCLES, or NS).
    stacked_cycles = []
    prefetch_data = plot_data.query(f'`is-random` == 1 & `prefetch-locality` > -1 & `prefetched-cache-lines` == {cache_lines}')
    if 'cycles' in metrics:
        if 'CYCLE_ACTIVITY.STALLS_TOTAL' in metrics:
            sequential['compute-cycles'] = sequential['cycles'] - sequential['CYCLE_ACTIVITY.STALLS_TOTAL']
            prefetch_data['overhead-cycles'] = prefetch_data['cycles'].apply(lambda cycles: cycles - sequential['compute-cycles'])
            prefetch_data['overhead-compute-cycles'] = prefetch_data['overhead-cycles'] - prefetch_data['CYCLE_ACTIVITY.STALLS_TOTAL']
            prefetch_data['stalled-cycles'] = prefetch_data['CYCLE_ACTIVITY.STALLS_TOTAL']

            if 'CYCLE_ACTIVITY.STALLS_MEM_ANY' in metrics:
                prefetch_data['stalled-cycles-mem'] = prefetch_data['CYCLE_ACTIVITY.STALLS_MEM_ANY']
                prefetch_data['stalled-cycles-other'] = prefetch_data['stalled-cycles'] - prefetch_data['stalled-cycles-mem']
                stacked_cycles = ['overhead-compute-cycles', 'stalled-cycles-mem', 'stalled-cycles-other']
            else:
                stacked_cycles = ['overhead-compute-cycles', 'stalled-cycles']
        else:
            prefetch_data['overhead-cycles'] = prefetch_data['cycles'].apply(lambda cycles: cycles - sequential['cycles'])
            stacked_cycles = ['overhead-cycles']
    else:
        prefetch_data['overhead-ns'] = prefetch_data['ns'].apply(lambda ns: ns - sequential['ns'])
        stacked_cycles = ['overhead-ns']

    ## Fetch the best prefetching runs per locality
    min_indicator = 'cycles' if 'cycles' in metrics else 'ns'
    min_indices = prefetch_data.groupby(['is-random', 'page-size', 'prefetch-locality'])[min_indicator].idxmin()
    prefetch_data = prefetch_data.loc[min_indices]

    if 'overhead-cycles' in prefetch_data.columns:
        max_y = prefetch_data['overhead-cycles'].max()
    else:
        max_y = prefetch_data['overhead-ns'].max()

    ## Sort by prefetch locality
    prefetch_data['prefetch-locality'] = pd.Categorical(prefetch_data['prefetch-locality'], categories=[0, 3, 1, 2], ordered=True)
    prefetch_data.sort_values(by='prefetch-locality', inplace=True)

    ## Sort by page size
    prefetch_data['page-size'] = pd.Categorical(prefetch_data['page-size'], categories=['4K', '2M', '1G'], ordered=True)
    unique_memory_page_sizes = prefetch_data['page-size'].unique()

    ## Group the data
    page_size_groups = prefetch_data.pivot(index='page-size', columns=['is-random', 'prefetch-locality'], values=stacked_cycles)
    print(page_size_groups.to_string())

    bar_width = 0.2
    x = np.arange(len(unique_memory_page_sizes))
    bottoms = {}
    labels_x = {}
    legend_bars = {}

    fig, ax = plt.subplots(figsize=(4.5,4.35))

    for j, (column, column_df) in enumerate(page_size_groups.items()):
        label = prefetch_locality_names[column[2]]
        fill_pattern = bar_fills[column[0]]
        color = colors[label]

        if label in bottoms:
            bottom = bottoms[label]
            bottoms[label] = bottom + column_df
        else:
            bottom = None
            bottoms[label] = column_df

        if label in labels_x:
            x_pos = labels_x[label]
        else:
            x_pos = j
            labels_x[label] = j

        edge_color = 'black'

        bar = ax.bar(.21 + x + (x_pos - 1.625 / 2) * bar_width, column_df, bar_width, bottom=bottom, label=label, color=color, hatch=fill_pattern, edgecolor=edge_color)
        if label not in legend_bars:
            legend_bars[label] = bar

    # Grid
    ax.spines[['right', 'top']].set_visible(False)
    ax.grid(True, color='#aaaaaa', linestyle='--', axis='y', linewidth=.4)
    ax.set_axisbelow(True)

    # Labels
    ax.set_xlabel('memory page size', fontsize=14)
    if 'cycles' in metrics:
        ax.set_ylabel('overhead (cycles / cache line)', fontsize=16)
    else:
        ax.set_ylabel('overhead (ns / cache line)', fontsize=16)

    # Ticks
    #ax.tick_params(labelsize=12, axis='y')
    ax.set_xticks(x + 5 * bar_width / len(plot_data['prefetch-locality'].unique()))
    ax.set_xticklabels(plot_data['page-size'].unique())
    #ax.set_yticks(list(range(0, int(max_y * 1.1), 5)))
    y_tick = 1 if max_y <= 5 else 2 if max_y <= 10 else 5 if max_y <= 50 else 10
    ax.set_yticks(list(range(0, int(max_y * 1.05), y_tick)))
    ax.yaxis.set_tick_params(labelsize=14)
    ax.xaxis.set_tick_params(labelsize=14)

    # Legend
    legend_handles = [handle for key, handle in legend_bars.items()]
    for metric in stacked_cycles:
        legend_handles.append(patches.Patch(edgecolor='black', fill=False, label=column_names[metric], hatch=bar_fills[metric]))
    plt.legend(loc='upper center', handles=legend_handles, ncol=2, fontsize=13, bbox_to_anchor=(.5, 1.35))

    # Arrow
    max_y = max(series.max() for series in bottoms.values())
    ax.annotate('better', xy=(2,max_y * .9), xytext=(2,max_y * .73), ha='center', fontsize=13, arrowprops={'edgecolor': 'black', 'facecolor': 'black', 'arrowstyle': '<|-', 'lw': 2})

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.7)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()


#############################################################
#   Single Samples                                          #
#                                                           #
#############################################################
def samples_per_cache_line(data, prefetched_cache_lines, file_path_out, is_small: bool, xticks: list = None, samples_output_file: str = None, samples_average_output_file: str = None):
    # Plotting the data
    fig, ax = plt.subplots(figsize=(4.75, 4))

    # Single samples
    if not type(data) is list:
        plt.scatter(data['in_block_offset'], data['weight'], color=colors['samples'], label='single samples', s=12)

    if samples_output_file is not None and samples_average_output_file is not None:
        samples_for_output = data[['in_block_offset', 'weight']]
        samples_for_output['in_block_offset'] = samples_for_output['in_block_offset'] + 1
        samples_for_output.to_csv(path_or_buf=samples_output_file, sep=' ', index=False)

        avg_samples_for_output = samples_for_output.groupby('in_block_offset')['weight'].mean().reset_index()
        avg_samples_for_output.to_csv(path_or_buf=samples_average_output_file, sep=' ', index=False)

    # Average
    if not type(data) is list:
        avg_latency_per_cache_line = data.groupby('in_block_offset')['weight'].mean().reset_index()
        ax.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], color=colors['prefetchNTA'], label='average latency', marker='o', linestyle='--', markersize=11)

        avg_latency_per_cache_line = data.groupby('in_block_offset')['weight'].median().reset_index()
        ax.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], color=colors['prefetchT0'], label='median latency', marker='x', linestyle='--', markersize=11)
    else:
        for item in data:
            name = item[0]
            samples = item[1]

            avg_latency_per_cache_line = samples.groupby('in_block_offset')['weight'].mean().reset_index()
            ax.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], color=colors[name], label=name, marker=markers[name], linestyle='--', markersize=11)

    if prefetched_cache_lines:
        plt.axvspan(0, float(prefetched_cache_lines) - .5, color=colors['prefetched-lines'], alpha=0.15)
        plt.text(float(prefetched_cache_lines) / 2, avg_latency_per_cache_line['weight'].max() * .25, 'software-prefetched', fontsize=14, ha='center', rotation=0)


    ax.spines[['right', 'top']].set_visible(False)
    ax.set_xlabel('cache line', fontsize=21 if is_small else 14)
    ax.set_ylabel('latency (cycles)', fontsize=16)
    ax.grid(True, color='#aaaaaa', linestyle='--', axis='y', linewidth=.4)

    # Axis
    if xticks is not None:
        ax.set_xticks(xticks)
    ax.xaxis.set_major_formatter(mticker.FuncFormatter(lambda x,pos: int(x + 1)))
    ax.xaxis.set_tick_params(labelsize=15)
    ax.yaxis.set_tick_params(labelsize=14)

    # Legend
    #plt.legend(loc='upper center', handles=legend_handles, ncol=2, fontsize=11, bbox_to_anchor=(.5, 1.35))
    plt.legend(loc='upper center', fontsize=13, ncol=2, bbox_to_anchor=(.5, 1.30 if type(data) is list else 1.18))

    # Arrow
    if type(data) is list:
        max_y = max([d[1].groupby('in_block_offset')['weight'].mean().reset_index()['weight'].max() for d in data])
    else:
        max_y = data['weight'].max()

    ax.annotate('better', xy=(2, max_y * .9), xytext=(2, max_y * .7), ha='center', fontsize=13, arrowprops={'edgecolor': 'black', 'facecolor': 'black', 'arrowstyle': '<|-', 'lw': 2})

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.75)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()

#############################################################
#   Tree                                                    #
#                                                           #
#############################################################
def tree(data, file_path_out):
    data.set_index('Machine', inplace=True)

    base_values_smt = data["Best without Coroutines (SMT)"]
    base_values_no_smt = data["Best without Coroutines"]

    speedups = pd.DataFrame()
    for column in data.columns:
        if "SMT" in column:
            speedups[column] = base_values_smt / data[column]
        else:
            speedups[column] = base_values_no_smt / data[column]

    speedups = speedups.drop(columns="Best without Coroutines (SMT)")
    speedups = speedups.drop(columns="Best without Coroutines")

    print(speedups.to_string())

    num_machines = len(speedups)
    num_columns = len(speedups.columns)
    bar_width = 0.15
    index = np.arange(num_machines)
    opacity = 0.8

    fig, ax = plt.subplots(figsize=(9, 5.95))

    plt.axhline(y=1.0, color='black', linestyle='--', linewidth=3)
    plt.text(num_machines - .2, 1.1, 'baseline', fontsize=20, ha='right', rotation=90)

    legend_handles = []
    for i, column in enumerate(speedups.columns):
        fill_pattern = '' if 'SMT' not in column else bar_fills['smt']

        handle = ax.bar(index + i * bar_width - .235, speedups[column], bar_width,
                alpha=opacity,
                color=colors[column],
                edgecolor='black',
                hatch=fill_pattern,
                label=column.replace(' size', ''))
        if "SMT" not in column:
            legend_handles.append(handle)

    plt.ylim(0, speedups.max().max() + 0.1)
    plt.xlabel('architecture', fontsize=22)
    plt.ylabel('speedup (over baseline)', fontsize=24)
    plt.xticks(index + bar_width, speedups.index, rotation=19, ha='center', fontsize=18)
    plt.yticks(fontsize=18)
    ax.yaxis.set_major_formatter(r'{x:1.1f}$\times$')

    # Legend
    legend_handles.append(patches.Patch(edgecolor='black', fill=False, label='single threaded', hatch=''))
    legend_handles.append(patches.Patch(edgecolor='black', fill=False, label='with SMT', hatch=bar_fills['smt']))

    legend_handles[1], legend_handles[2] = legend_handles[2], legend_handles[1]
    legend_handles[1], legend_handles[3] = legend_handles[3], legend_handles[1]
    legend_handles[3], legend_handles[4] = legend_handles[4], legend_handles[3]

    ax.spines[['right', 'top']].set_visible(False)
    plt.legend(loc='upper center', handles=legend_handles, fontsize=18, ncols=3, bbox_to_anchor=(.5, 1.27))
    plt.grid(True, color='#aaaaaa', linestyle='--', axis='y', linewidth=.4)

    # Arrow
    max_y = speedups.max().max()
    plt.annotate('better', xy=(-.35, max_y * .87), xytext=(-.35, max_y * .99), ha='center', fontsize=16, arrowprops={'edgecolor': 'black', 'facecolor': 'black', 'arrowstyle': '<|-', 'lw': 2})

    plt.subplots_adjust(bottom=0.15, top=0.7)
    plt.tight_layout()
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()


#############################################################
#   Make the Plots happen                                   #
#                                                           #
#############################################################
parser = argparse.ArgumentParser()
parser.add_argument('--in', dest='in_dir', help='Dir with raw data files', required=True)
parser.add_argument('--out', dest='out_dir', help='Folder to save plots.', required=True)
args = parser.parse_args()


#######################     Trend       #######################
cache_lines = 1
intel = [
    (2012, 'Sandy Bridge', get_trend_prefetching_improvement(f'{args.in_dir}/Intel/sandy-bridge/out.csv'), 0, -.07),
    (2013, 'Haswell', get_trend_prefetching_improvement(f'{args.in_dir}/Intel/haswell/out.csv'), 0, .015),
    (2015, 'Skylake', get_trend_prefetching_improvement(f'{args.in_dir}/Intel/skylake/out.csv'), 0, .02),
    (2019, 'Cascade Lake', get_trend_prefetching_improvement(f'{args.in_dir}/Intel/cascade-lake-arnor/out.csv'), 0, -.07),
    (2023, 'Sapphire Rapids', get_trend_prefetching_improvement(f'{args.in_dir}/Intel/sapphire-rapids/out.csv'), 0, .025)
]

amd = [
    (2017.25, 'Zen 1', get_trend_prefetching_improvement(f'{args.in_dir}/AMD/zen1-desktop/out.csv'), 0, .03),
    (2019, 'Zen 2', get_trend_prefetching_improvement(f'{args.in_dir}/AMD/zen2-desktop/out.csv'), -0.1, .02),
    (2022, 'Zen 4', get_trend_prefetching_improvement(f'{args.in_dir}/AMD/zen4-desktop/out.csv'), 0, .015),
]

arm = [
    (2012, 'Cortex-A53', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/cortex-a53/out.csv'), 0, -.075),
    (2014.75, 'Cortex-A72', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/cortex-a72/out.csv'), .08, -.1),
    (2016, 'Cortex-A73', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/cortex-a73/out.csv'), 0, -.07),
    (2018, 'Cortex-A76', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/cortex-a76/out.csv'), 0, .025),
]

arm_server = [
    (2018, 'Thunder X2', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/cavium-thunderx2/out_trend.csv'), 0, -.08),
    (2020, 'Altra', get_trend_prefetching_improvement(f'{args.in_dir}/ARM/ampere-altra/out.csv'), 0.75, -.0275),
]

trend({ 'Intel': (intel, colors['Intel'], markers['Intel']), 'ARM': (arm, colors['ARM'], markers['ARM']), 'AMD': (amd, colors['AMD'], markers['AMD']), 'ARM Server': (arm_server, colors['ARM Server'], markers['ARM Server'])}, args.out_dir)

####################### Stall Breakdown     #######################
for machine in ['cascade-lake-gondor', 'sapphire-rapids-umbar']:
    print(f'### {machine} >')
    machine_data = pd.read_csv(os.path.join(args.in_dir, 'Intel', machine, 'out.csv'))
    machine_data = machine_data.query('`load-heavy` == 0')

    if 'EXE_ACTIVITY.BOUND_ON_LOADS' in machine_data.columns and 'EXE_ACTIVITY.BOUND_ON_STORES' in machine_data.columns:
        machine_data['CYCLE_ACTIVITY.STALLS_MEM_ANY'] = machine_data['EXE_ACTIVITY.BOUND_ON_LOADS'] + machine_data['EXE_ACTIVITY.BOUND_ON_STORES']

    machine_name = machine.split('-')[-1]
    suffix = '' if machine_name == 'gondor' else f'-{machine_name}'

    for cache_lines in [1, 16]:
        print(f'# {cache_lines} cache lines')
        stalls_breakdown(
            machine_data,
            os.path.join(args.out_dir, f'cost-cycle-breakdown-{cache_lines}-cache-lines{suffix}.pdf'),
            ['cycles', 'CYCLE_ACTIVITY.STALLS_TOTAL', 'CYCLE_ACTIVITY.STALLS_MEM_ANY'],
            1,
            cache_lines,
            False
        )

    print(f'< {machine} ´###')

####################### Samples      #######################
gondor_samples = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'nta-16-1g.csv')).query(f'`in_line_offset` == 0 & `is_workload` == 1')
samples_per_cache_line(gondor_samples, None, os.path.join(args.out_dir, f'nta-16-1g-pref.pdf'), False, samples_output_file=os.path.join(args.out_dir, f'nta-16-1g-pref.dat'), samples_average_output_file=os.path.join(args.out_dir, f'nta-16-1g-pref-avg.dat'))

gondor_samples_random = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-random-block-16-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')
gondor_samples_nta_8 = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-locality-0-block-16-pref-8-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')
gondor_samples_nta_10 = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-locality-0-block-16-pref-10-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')
gondor_samples_nta_12 = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-locality-0-block-16-pref-12-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')

samples_per_cache_line(
    [('random', gondor_samples_random), ('sw-pref. 8 lines', gondor_samples_nta_8), ('sw-pref. 10 lines', gondor_samples_nta_10), ('sw-pref. 12 lines', gondor_samples_nta_12)],
    None,
       os.path.join(args.out_dir, 'nta-0-8-10-12.pdf'),
    False,
    [0] + list(range(1, 16, 2))
)

gondor_samples_t0_10 = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-locality-3-block-16-pref-10-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')
gondor_samples_t1_10 = pd.read_csv(os.path.join(args.in_dir, 'samples', 'Intel', 'cascade-lake-gondor', 'sample-locality-2-block-16-pref-10-l1-l2-l3.csv')).query(f'`in_line_offset` == 4 & `is_workload` == 1')

samples_per_cache_line(
    [('prefetchT0', gondor_samples_t0_10), ('prefetchT1', gondor_samples_t1_10), ('prefetchNTA', gondor_samples_nta_10)],
    10,
    os.path.join(args.out_dir, 'temporal-vs-nta-10.pdf'),
    False,
    [0, 5, 9, 11, 15]
)

####################### Tree      #######################
tree_data = pd.read_csv(os.path.join(args.in_dir, 'tree.csv'))
tree(tree_data, os.path.join(args.out_dir, 'tree-1-core.pdf'))