import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os
import argparse
import matplotlib.patches as patches
import math

plt.rcParams.update({
    "font.family": "serif",
    "text.usetex": True,
    'text.latex.preamble': "\\usepackage{libertine}\\usepackage[libertine]{newtxmath}",
})

prefetch_locality_names = {-1: 'Sequential', 0: 'NTA', 1: 'T2', 2: 'T1', 3: 'T0'}
column_names = {
    'overhead-compute-cycles': 'compute cycles',
    'stalled-cycles-mem': 'memory stalled cycles',
    'stalled-cycles-other': 'other stalled cycles',
    'stalled-cycles': 'stalled cycles',
    'overhead-cycles': 'cycles',
    'overhead-ns': 'nanoseconds'
}

fill_mapping = {
    'overhead-compute-cycles': '',
    'stalled-cycles-mem': '//',
    'stalled-cycles-other': '...',
    'stalled-cycles': '...',
    'overhead-cycles': '',
    'overhead-ns': ''
}

color_mapping = {
    'NTA': '#778da9',
    'T2': '#e07a5f',
    'T1': '#81b29a',
    'T0': '#f2cc8f',
    'Full': '#39568C',
    'Full': '#39568C',
    'Adjacent': '#73D055',
    'Sequential': '#588157',
    'Random': '#c1121f'
}


def plot_sequential_prefetch_overhead(data, file_path_out: str, metric: str, hyper_threads: int):
    sequential_runs = data if 'hyper-threads' not in data.columns else data.query(f'`hyper-threads` == {hyper_threads}')
    if 'is-interleaved' in data.columns:
        sequential_runs = sequential_runs.query('`is-interleaved` == 0')
    sequential_runs = sequential_runs.query('`is-random` == 0 & `page-size` == "4K"')

    unique_cache_lines = sequential_runs['cache-lines'].unique()
    num_rows = int(math.ceil(len(unique_cache_lines) / 3))

    plt.figure(figsize=(10, 12))
    for i, cache_line in enumerate(unique_cache_lines, 1):

        ax = plt.subplot(num_rows, 3, i)

        cache_line_runs = sequential_runs.query(f'`cache-lines` == {cache_line}')

        sequential_without_prefetching = cache_line_runs.query('`prefetch-locality` < 0').iloc[0][metric]

        for locality in cache_line_runs.query('`prefetch-locality` >= 0')['prefetch-locality'].unique():
            locality_name = prefetch_locality_names[locality]
            sequential_with_prefetching = cache_line_runs.query(f'`prefetch-locality` == {locality}').iloc[0][metric]
            diff = sequential_with_prefetching - sequential_without_prefetching
            ax.bar(locality_name, diff)

        ax.set_title(f'block = {cache_line} cache lines', fontsize=11)

    plt.xlabel('')
    plt.ylabel(f'overhead to sequential ({metric})')
    plt.tight_layout()
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()


def plot_random_prefech_overhead(data, file_path_out: str,  metric: str, is_max_better: bool, hyper_threads: int):
    hyper_threads_data = data if 'hyper-threads' not in data.columns else data.query(f'`hyper-threads` == {hyper_threads}')
    hyper_threads_data = hyper_threads_data.query('`cache-lines` == 1')

    plot_data = {'page-size': [], 'prefetch-locality': [], metric: []}

    for memory_page_size in hyper_threads_data['page-size'].unique():
        memory_page_data = hyper_threads_data.query(f'`page-size` == "{memory_page_size}"')

        sequential_without_prefetching = memory_page_data.query('`is-random` == 0 & `prefetch-locality` < 0').iloc[0][metric]

        for prefetch_locality in [p for p in memory_page_data['prefetch-locality'].unique() if p >= 0]:
            prefetch = memory_page_data.query(f'`is-random` == 1 & `prefetch-locality` == {prefetch_locality}')[metric]
            prefetch = prefetch.max() if is_max_better else prefetch.min()
            diff = prefetch - sequential_without_prefetching

            plot_data['page-size'].append(memory_page_size)
            plot_data['prefetch-locality'].append(prefetch_locality_names[prefetch_locality])
            plot_data[metric].append(diff)


    # Group by 'page-size' and 'prefetch-locality' and find the minimum cycle count for each group
    plot_data = pd.DataFrame(plot_data)

    # Order data
    plot_data['page-size'] = pd.Categorical(plot_data['page-size'], categories=['4K', '2M', '1G'], ordered=True)
    best_results = pd.DataFrame(plot_data).groupby(['page-size', 'prefetch-locality'])[metric]
    if is_max_better:
        best_results = best_results.max().reset_index()
    else:
        best_results = best_results.min().reset_index()
    pivot_data_swapped = best_results.pivot(index='page-size', columns='prefetch-locality', values=metric)

    pivot_data_swapped.plot(kind='bar', figsize=(5, 3))
    plt.xlabel('memory page size')
    plt.ylabel(f'overhead to sequential ({metric})')
    plt.legend(title='Prefetch Locality', bbox_to_anchor=(1.05, 1), loc='upper left')
    plt.grid(axis='y')
    plt.tight_layout()
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()


'''
    If MEM_STALLS and TOTAL_STALLS:
        compute-cycles = seq.cycles - seq.total_stalls
        [overhead-cycles] = cycles - compute-cycles
        [overhead-compute-cycles] = overhead-cycles - total_stalls
        [overhead-mem-stalls] = mem_stalls
        [overhead-other-stalls] = total_stalls - mem_stalls
    elif TOTAL_STALLS:
        compute-cycles = seq.cycles - seq.total_stalls
        [overhead-cycles] = cycles - compute-cycles
        [overhead-compute-cycles] = overhead-cycles - total_stalls
        [overhead-stalls] = total_stalls
    elif cycles:
        compute-cycles = seq.cycles
        [overhead-cycles] = cycles - compute-cycles
    else:
        compute-ns = seq.ns
        [overhead-ns] = ns - compute-ns
'''
def stalls_breakdown(data, file_path_out: str, metrics: list, hyper_threads: int, cache_lines: int, is_interleaved: int):

    plot_data = data.query(f'`cache-lines` == {cache_lines}')
    if 'hyper-threads' in data.columns:
        plot_data = plot_data.query(f'`hyper-threads` == {hyper_threads}')

    if 'is-interleaved-prefetching' in data.columns:
        plot_data = plot_data.query(f'`is-interleaved-prefetching` == {is_interleaved}')

    ## Sequential run to compare prefetched runs with
    sequential = plot_data.query('`is-random` == 0 & `prefetch-locality` == -1').iloc[0]

    ## Calculate overhead columns based in the available metrics (STALLS + MEM_STALLS, STALLS, CYCLES, or NS).
    stacked_cycles = []
    prefetch_data = plot_data.query(f'`is-random` == 1 & `prefetch-locality` > -1 & `prefetched-cache-lines` == {cache_lines}')
    if 'cycles' in metrics:
        if 'CYCLE_ACTIVITY.STALLS_TOTAL' in metrics:
            sequential['compute-cycles'] = sequential['cycles'] - sequential['CYCLE_ACTIVITY.STALLS_TOTAL']
            prefetch_data['overhead-cycles'] = prefetch_data['cycles'].apply(lambda cycles: cycles - sequential['compute-cycles'])
            prefetch_data['overhead-compute-cycles'] = prefetch_data['overhead-cycles'] - prefetch_data['CYCLE_ACTIVITY.STALLS_TOTAL']
            prefetch_data['stalled-cycles'] = prefetch_data['CYCLE_ACTIVITY.STALLS_TOTAL']

            if 'CYCLE_ACTIVITY.STALLS_MEM_ANY' in metrics:
                prefetch_data['stalled-cycles-mem'] = prefetch_data['CYCLE_ACTIVITY.STALLS_MEM_ANY']
                prefetch_data['stalled-cycles-other'] = prefetch_data['stalled-cycles'] - prefetch_data['stalled-cycles-mem']
                stacked_cycles = ['overhead-compute-cycles', 'stalled-cycles-mem', 'stalled-cycles-other']
            else:
                stacked_cycles = ['overhead-compute-cycles', 'stalled-cycles']
        else:
            prefetch_data['overhead-cycles'] = prefetch_data['cycles'].apply(lambda cycles: cycles - sequential['cycles'])
            stacked_cycles = ['overhead-cycles']
    else:
        prefetch_data['overhead-ns'] = prefetch_data['ns'].apply(lambda ns: ns - sequential['ns'])
        stacked_cycles = ['overhead-ns']

    ## Fetch the best prefetching runs per locality
    min_indicator = 'cycles' if 'cycles' in metrics else 'ns'
    min_indices = prefetch_data.groupby(['is-random', 'page-size', 'prefetch-locality'])[min_indicator].idxmin()
    prefetch_data = prefetch_data.loc[min_indices]

    if 'overhead-cycles' in prefetch_data.columns:
        max_y = prefetch_data['overhead-cycles'].max()
    else:
        max_y = prefetch_data['overhead-ns'].max()

    ## Sort by page size
    prefetch_data['page-size'] = pd.Categorical(prefetch_data['page-size'], categories=['4K', '2M', '1G'], ordered=True)
    unique_memory_page_sizes = prefetch_data['page-size'].unique()

    ## Group the data
    page_size_groups = prefetch_data.pivot(index='page-size', columns=['is-random', 'prefetch-locality'], values=stacked_cycles)

    bar_width = 0.1
    x = np.arange(len(unique_memory_page_sizes))
    bottoms = {}
    labels_x = {}
    legend_bars = {}

    fig, ax = plt.subplots(figsize=(4.5,3))

    for j, (column, column_df) in enumerate(page_size_groups.items()):
        label = prefetch_locality_names[column[2]]
        fill_pattern = fill_mapping[column[0]]
        color = color_mapping[label]

        if label in bottoms:
            bottom = bottoms[label]
            bottoms[label] = bottom + column_df
        else:
            bottom = None
            bottoms[label] = column_df

        if label in labels_x:
            x_pos = labels_x[label]
        else:
            x_pos = j
            labels_x[label] = j

        bar = ax.bar(.1 + x + (x_pos - 1.625 / 2) * bar_width, column_df, bar_width, bottom=bottom, label=label, color=color, hatch=fill_pattern, edgecolor='black')
        if label not in legend_bars:
            legend_bars[label] = bar

    # Grid
    ax.grid(color='gray', linestyle='-', linewidth=.4, axis='y')
    ax.set_axisbelow(True)

    # Labels
    ax.set_xlabel('memory page size', fontsize=14)
    if 'cycles' in metrics:
        ax.set_ylabel('overhead (cycles / cache line)', fontsize=14)
    else:
        ax.set_ylabel('overhead (ns / cache line)', fontsize=14)

    # Ticks
    #ax.tick_params(labelsize=12, axis='y')
    ax.set_xticks(x + 5 * bar_width / len(plot_data['prefetch-locality'].unique()))
    ax.set_xticklabels(plot_data['page-size'].unique())
    #ax.set_yticks(list(range(0, int(max_y * 1.1), 5)))
    y_tick = 1 if max_y <= 5 else 2 if max_y <= 10 else 5 if max_y <= 100 else 10
    ax.set_yticks(list(range(0, int(max_y * 1.05), y_tick)))

    # Legend
    legend_handles = [handle for key, handle in legend_bars.items()]
    for metric in stacked_cycles:
        legend_handles.append(patches.Patch(edgecolor='black', fill=False, label=column_names[metric], hatch=fill_mapping[metric]))
    plt.legend(loc='upper right', handles=legend_handles, ncol=2, fontsize=11) #

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.85)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()


parser = argparse.ArgumentParser()
parser.add_argument('--in', dest='in_file', help='CSV file with data.', required=True)
parser.add_argument('--out', dest='out_dir', help='Folder to save plots (default will be the folder of the in-file).', default=None)

args = parser.parse_args()
out = args.out_dir if args.out_dir else os.path.dirname(args.in_file)
data = pd.read_csv(args.in_file)

if 'page-size' not in data.columns:
    data['page-size'] = '4K'

if 'load-heavy' in data.columns:
    data = data.query('`load-heavy` == 0')

#if 'cycles' in data.columns:
    #plot_sequential_prefetch_overhead(data, os.path.join(out, 'seq-overhead-cycles.pdf'), 'cycles', 1)
    #plot_random_prefech_overhead(data, os.path.join(out, 'seq-vs-random-overhead-cycles.pdf'), 'cycles', False, 1)
#else:
    #plot_sequential_prefetch_overhead(data, os.path.join(out, 'seq-overhead-ns.pdf'), 'ns', 1)
    #plot_random_prefech_overhead(data, os.path.join(out, 'seq-vs-random-overhead-ns.pdf'), 'ns', False, 1)

interleaves = [0] #if 'is-interleaved-prefetching' not in data.columns else [0, 1]
for cache_lines in data['cache-lines'].unique():
    cache_line_suffix = f'{cache_lines}-cache-lines'
    for is_interleaved in interleaves:
        suffix = cache_line_suffix if not is_interleaved else f'{cache_line_suffix}-interleaved'
        if 'CYCLE_ACTIVITY.STALLS_MEM_ANY' in data.columns and 'CYCLE_ACTIVITY.STALLS_TOTAL' in data.columns:
            stalls_breakdown(data,  os.path.join(out, f'cost-cycle-breakdown-{suffix}.pdf'), ['cycles', 'CYCLE_ACTIVITY.STALLS_TOTAL', 'CYCLE_ACTIVITY.STALLS_MEM_ANY'], 1, cache_lines, is_interleaved)
        if 'CYCLE_ACTIVITY.STALLS_TOTAL' in data.columns:
            stalls_breakdown(data,  os.path.join(out, f'cost-cycle-stalls-{suffix}.pdf'), ['cycles', 'CYCLE_ACTIVITY.STALLS_TOTAL'], 1, cache_lines, is_interleaved)
        if 'cycles' in data.columns:
            stalls_breakdown(data,  os.path.join(out, f'cost-cycles-{suffix}.pdf'), ['cycles'], 1, cache_lines, is_interleaved)
        stalls_breakdown(data,  os.path.join(out, f'cost-ns-{suffix}.pdf'), ['ns'], 1, cache_lines, is_interleaved)