import pandas as pd
import matplotlib.pyplot as plt
import argparse
import os
import re

plt.rcParams.update({
    "font.family": "serif",
    "text.usetex": True,
    'text.latex.preamble': "\\usepackage{libertine}\\usepackage[libertine]{newtxmath}",
})


def plot_latency_per_cacheline(data, prefetched_cache_lines, file_path_out):
    # Plotting the data
    plt.figure(figsize=(10, 6))

    # Filtering data based on memory type and plotting each with a different color
    for mem_type, color, label in zip(['is_mem_l1', 'is_mem_l2', 'is_mem_l3', 'is_mem_lfb', 'is_mem_ram'],
                                      ['#778da9', '#81b29a', '#f2cc8f', '#c1121f', '#e07a5f'],
                                      ['L1', 'L2', 'L3', 'LFB', 'RAM']):
        filtered_data = data[data[mem_type] == 1]
        plt.scatter(filtered_data['in_block_offset'], filtered_data['weight'], color=color, label=label, s=10)

    plt.xlabel('Cache Line')
    plt.ylabel('Latency')
    plt.title('Latency per Cache Line')
    plt.legend()
    plt.grid(True)

    avg_latency_per_cache_line = data.groupby('in_block_offset')['weight'].mean().reset_index()
    plt.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], color='black', label='Average Latency', marker='o')
    for idx, row in avg_latency_per_cache_line.iterrows():
        plt.text(row['in_block_offset'], row['weight'], f"{row['weight']:.2f}", color='black', ha='right', va='bottom')

    if prefetched_cache_lines:
        plt.axvspan(0, prefetched_cache_lines - 1, color='#cccccc', alpha=0.3)

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.85)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')
    plt.close()



def scan_folder_create_dataframe(folder_path):
    # Initialize an empty list to store the rows of the DataFrame
    data = []

    # Iterate through each file in the folder
    for filename in os.listdir(folder_path):
        if filename.startswith("sample-") and filename.endswith(".csv"):
            # Initialize the dictionary to store file information
            file_info = {
                'locality': -1,
                'is-random': 0,
                'is-sequential': 0,
                'block': None,
                'pref': 0,  # Default to 0 for 'random' and 'sequential'
                'is-l1-pref': 0,
                'is-l2-pref': 0,
                'is-l3-pref': 0,
                'filename': os.path.join(folder_path, filename)
            }

            # Extract the locality
            locality_match = re.search(r'locality-(\d+)', filename)
            if locality_match:
                file_info['locality'] = int(locality_match.group(1))

            # Check for 'random' or 'sequential'
            if 'random' in filename:
                file_info['is-random'] = 1
            elif 'sequential' in filename:
                file_info['is-sequential'] = 1

            # Extract the block number
            block_match = re.search(r'block-(\d+)', filename)
            if block_match:
                file_info['block'] = int(block_match.group(1))

            # Extract the pref value if not 'random' or 'sequential'
            if file_info['is-random'] == 0 and file_info['is-sequential'] == 0:
                pref_match = re.search(r'pref-(\d+)', filename)
                if pref_match:
                    file_info['pref'] = int(pref_match.group(1))

            # Check for 'l1', 'l2', 'l3'
            if 'l1' in filename:
                file_info['is-l1-pref'] = 1
            if 'l2' in filename:
                file_info['is-l2-pref'] = 1
            if 'l3' in filename:
                file_info['is-l3-pref'] = 1

            # Add the file info to the data list
            data.append(file_info)

    # Create a DataFrame from the data list
    return pd.DataFrame(data)

def create_file_mapping(all_files, filter: str):

    locality_names = {0: 'NTA', 1: 'T2', 2: 'T1', 3: 'T0'}

    filtered_files = all_files if not filter else all_files.query(filter)

    mapping = {}

    for index, row in filtered_files.iterrows():
        name = ''
        if row['is-random'] == 1:
            name = 'Random'
        elif row['is-sequential'] == 1:
            name = 'Sequential'
        else:
            name = locality_names[int(row['locality'])]

        if row['is-l1-pref']:
            name += ' + L1'
        if row['is-l2-pref']:
            name += ' + L2'
        if row['is-l3-pref']:
            name += ' + LLC'

        mapping[name] = row['filename']

    return mapping

def plot(files: dict, file_path_out, offset : int, prefetched_cache_lines = None):

    # Plotting the data
    plt.figure(figsize=(4, 2.5))

    x_ticks = (9999999999,0)
    max_y_tick = 0

    for name, file_name in files.items():
        data = pd.read_csv(file_name).query(f'`in_line_offset` == {offset} & `is_workload` == 1')

        avg_latency_per_cache_line = data.groupby('in_block_offset')['weight'].mean().reset_index()
        plt.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], label=name, marker='o', linestyle='--')

        x_ticks = (min(avg_latency_per_cache_line['in_block_offset'].min(), x_ticks[0]), max(avg_latency_per_cache_line['in_block_offset'].max(), x_ticks[1]))
        max_y_tick = max(max_y_tick, avg_latency_per_cache_line['weight'].max())

    if prefetched_cache_lines:
        plt.axvspan(0, float(prefetched_cache_lines) - .5, color='#bbbbbb', alpha=0.3)

    plt.xlabel('cache line index')
    plt.xticks(range(int(x_ticks[0]), int(x_ticks[1]) + 1))
    plt.ylim(0, max_y_tick * 1.05)
    plt.ylabel('latency (cpu cycles)')
    plt.legend()
    plt.grid(True, color='#cccccc')

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.85)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    plt.close()
    print(f'Created {file_path_out}')


parser = argparse.ArgumentParser()
parser.add_argument('--in', dest='in_dir', help='Dir with data files.', required=True)
parser.add_argument('--out', dest='out_dir', help='Folder to save plots (default will be the folder of the in-file).', default=None)
parser.add_argument('--only-custom', help='If set, will only plot custom plots', action='store_true')
parser.add_argument('--byte', type=int, help='Byte to sample (0 = first value and prefetches, 4 = second value, ...)', default=4)
parser.add_argument('--cache-lines', type=int, help='Block size to use for custom plots', default=16)

args = parser.parse_args()
out_dir = args.out_dir if args.out_dir else os.path.dirname(args.in_dir)

if not args.only_custom:
    # Auto-generated plots
    files = scan_folder_create_dataframe(args.in_dir)
    plot(create_file_mapping(files, '`is-random` == 1'), os.path.join(out_dir, 'random-average-latency.pdf'), args.byte)
    plot(create_file_mapping(files, '`is-sequential` == 1'), os.path.join(out_dir, 'sequential-average-latency.pdf'), args.byte)

    for prefetched in files['pref'].unique():
        if prefetched > 0:
            plot(create_file_mapping(files, f'`locality` == 0 & `pref` == {prefetched}'), os.path.join(out_dir, f'nta-pref-{prefetched}-average-latency.pdf'), args.byte, prefetched)
            #plot(create_file_mapping(files, f'`locality` == 1 & `pref` == {prefetched}'), os.path.join(out_dir, f't2-pref-{prefetched}-average-latency.pdf'), prefetched)
            plot(create_file_mapping(files, f'`locality` == 2 & `pref` == {prefetched}'), os.path.join(out_dir, f't1-pref-{prefetched}-average-latency.pdf'), args.byte, prefetched)
            plot(create_file_mapping(files, f'`locality` == 3 & `pref` == {prefetched}'), os.path.join(out_dir, f't0-pref-{prefetched}-average-latency.pdf'), args.byte, prefetched)

# Custom plots

## T0 for different prefetched lines
plot(
    {
        'random': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}-l1-l2-l3.csv'),
        'sw-pref. 8 lines': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-8-l1-l2-l3.csv'),
        'sw-pref. 10 lines': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-10-l1-l2-l3.csv'),
        'sw-pref. 12 lines': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-12-l1-l2-l3.csv'),
    },
    os.path.join(out_dir, 't0-0-8-10-12.pdf'),
    args.byte,
)

## NTA for different prefetched lines
plot(
    {
        'random': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}-l1-l2-l3.csv'),
        'sw-pref. 8 lines': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-8-l1-l2-l3.csv'),
        'sw-pref. 10 lines': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-10-l1-l2-l3.csv'),
        'sw-pref. 12 lines': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-12-l1-l2-l3.csv'),
    },
    os.path.join(out_dir, 'nta-0-8-10-12.pdf'),
    args.byte,
)

## Temporal vs NonTemporal for different prefetched lines
for prefetched_cache_lines in [8,10,12]:
    plot(
        {
            f'sw-pref. {prefetched_cache_lines} lines (T0)': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l2-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (T1)': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l2-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (NTA)': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l2-l3.csv'),
        },
        os.path.join(out_dir, f'temporal-vs-nta-{prefetched_cache_lines}.pdf'),
        args.byte,
        prefetched_cache_lines
    )

    plot(
        {
            f'sw-pref. {prefetched_cache_lines} lines (T0)': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l2-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (T1)': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l2-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (NTA)': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l2-l3.csv'),
        },
        os.path.join(out_dir, f'temporal-vs-nta-{prefetched_cache_lines}-only-l2.pdf'),
        args.byte,
        prefetched_cache_lines
    )

    plot(
        {
            f'sw-pref. {prefetched_cache_lines} lines (T0)': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (T1)': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l3.csv'),
            f'sw-pref. {prefetched_cache_lines} lines (NTA)': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l3.csv'),
        },
        os.path.join(out_dir, f'temporal-vs-nta-{prefetched_cache_lines}-only-l1.pdf'),
        args.byte,
        prefetched_cache_lines
    )

## Hardware Prefetchers (None, L2, L1+L2); Complete Random
plot(
    {
        'random without HWPF': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}.csv'),
        'random + L2 Pref': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}-l2-l3.csv'),
        'random + L1 Pref': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}-l1-l3.csv'),
        'random + HWPF': os.path.join(args.in_dir, f'sample-random-block-{args.cache_lines}-l1-l2-l3.csv'),
     },
    os.path.join(out_dir, 'random-hardware-prefetcher.pdf'),
    args.byte,
)

## Software Prefetches (NTA, T1, T0); Prefetching All CLs
plot(
    {
        'prefetchNTA': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{args.cache_lines}-l3.csv'),
        'prefetchT1': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{args.cache_lines}-l3.csv'),
        'prefetchT0': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-{args.cache_lines}-l3.csv'),
    },
    os.path.join(out_dir, 'software-prefetcher.pdf'),
    args.byte,
)

## Software Prefetches + Hardware Prefetches (T1 + L2, T1 + L1 + L2); Prefetching All CLs
plot(
    {
        'prefetchT1 + HWPF (L2)': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{args.cache_lines}-l2-l3.csv'),
        'prefetchT1 + HWPF (L1 + L2)': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{args.cache_lines}-l1-l2-l3.csv'),
        'prefetchT0 without HWPF': os.path.join(args.in_dir, f'sample-locality-3-block-{args.cache_lines}-pref-{args.cache_lines}-l3.csv'),
        'prefetchNTA without HWPF': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{args.cache_lines}-l3.csv'),
    },
    os.path.join(out_dir, f'sw-hw-interaction-{args.cache_lines}-{args.cache_lines}.pdf'),
    args.byte,
)

## Software Prefetches + Hardware Prefetches (T1 + L2, T1 + L1 + L2); Prefetching 14 CLs
for prefetched_cache_lines in [6, 8, 10,12]:
    plot(
        {
            'prefetchT1 without HWPF': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{prefetched_cache_lines}-interleaved-l3.csv'),
            'prefetchT1 + HWPF': os.path.join(args.in_dir, f'sample-locality-2-block-{args.cache_lines}-pref-{prefetched_cache_lines}-interleaved-l1-l2-l3.csv'),
            'prefetchNTA without HWPF': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{prefetched_cache_lines}-interleaved-l3.csv'),
            'prefetchNTA + HWPF': os.path.join(args.in_dir, f'sample-locality-0-block-{args.cache_lines}-pref-{prefetched_cache_lines}-l1-l2-l3.csv'),
            #'prefetchT0 without HWPF': os.path.join(args.in_dir, f'sample-locality-3-block-16-pref-{prefetched_cache_lines}-l3.csv'),
            #'prefetchT0 + HWPF': os.path.join(args.in_dir, f'sample-locality-3-block-16-pref-{prefetched_cache_lines}-l1-l2-l3.csv'),
        },
        os.path.join(out_dir, f'sw-hw-interaction-{args.cache_lines}-{prefetched_cache_lines}.pdf'),
        args.byte,
        prefetched_cache_lines
    )