import pandas as pd
import matplotlib.pyplot as plt
import argparse
import os

def plot_latency_per_cacheline(data, prefetched_cache_lines, file_path_out):
    # Plotting the data
    plt.figure(figsize=(5, 3))

    # Filtering data based on memory type and plotting each with a different color
    for mem_type, color, label in zip(['is_mem_l1', 'is_mem_l2', 'is_mem_l3', 'is_mem_lfb', 'is_mem_ram'],
                                      ['#778da9', '#81b29a', '#f2cc8f', '#c1121f', '#e07a5f'],
                                      ['L1', 'L2', 'L3', 'LFB', 'RAM']):
        filtered_data = data[data[mem_type] == 1]
        plt.scatter(filtered_data['in_block_offset'], filtered_data['weight'], color=color, label=label, s=10)

    plt.xlabel('cache Line', fontsize=14)
    plt.ylabel('latency (cycles)', fontsize=14)
    plt.legend(fontsize=11)
    plt.grid(color='gray', linestyle='-', linewidth=.4, axis='y')

    avg_latency_per_cache_line = data.groupby('in_block_offset')['weight'].mean().reset_index()
    plt.plot(avg_latency_per_cache_line['in_block_offset'], avg_latency_per_cache_line['weight'], color='black', label='Average Latency', marker='o', linestyle='--')
    for idx, row in avg_latency_per_cache_line.iterrows():
        plt.text(row['in_block_offset'], row['weight'], f"{row['weight']:.2f}", color='black', ha='right', va='bottom')

    if prefetched_cache_lines:
        plt.axvspan(0, prefetched_cache_lines - 1, color='#cccccc', alpha=0.3)

    # Adjust
    plt.subplots_adjust(bottom=0.15, top=0.85)
    plt.tight_layout()

    # Save plot
    plt.savefig(file_path_out)
    print(f'Created {file_path_out}')

parser = argparse.ArgumentParser()
parser.add_argument('--in', dest='in_file', help='File to plot', required=True)
parser.add_argument('--out', dest='out_dir', help='Folder to save plots (default will be the folder of the in-file).', default=None)
parser.add_argument('-k', help='Keep file name.', action='store_true')
parser.add_argument('--prefetched', type=int, help='Number of prefetched cache lines (will be highlighted)', default=None)
parser.add_argument('--byte', type=int, help='Byte to sample (0 = first value and prefetches, 4 = second value, ...)', default=4)

args = parser.parse_args()

data = pd.read_csv(args.in_file).query(f'`in_line_offset` == {args.byte} & `is_workload` == 1')
out_dir = args.out_dir if args.out_dir else os.path.dirname(args.in_file)

out_file = args.in_file.replace('.csv', '.pdf') if args.k else os.path.join(out_dir, f'latency-per-cacheline.pdf')
plot_latency_per_cacheline(data, args.prefetched, out_file)