import pandas as pd
import matplotlib.pyplot as plt
import os

plt.rcParams.update({
    "font.family": "serif",
    "text.usetex": True,
    'text.latex.preamble': "\\usepackage{libertine}\\usepackage[libertine]{newtxmath}",
})

def get_prefetching_improvement(file_path, block_size):
    data = pd.read_csv(file_path)
    random_condition = (data["is-random"] == 1) & (data["prefetch-distance"] == 0) & (data["cache-lines"] == block_size)
    if "hyper-threads" in data.columns:
        random_condition = random_condition & (data["hyper-threads"] == 1)
    random_max = data.loc[random_condition, "GB / second"].max()
    
    prefetching_condition = (data["is-random"] == 1) & (data["prefetch-distance"] > 0) & (data["cache-lines"] == block_size) & (data["prefetched-cache-lines"] == block_size)
    if "hyper-threads" in data.columns:
        prefetching_condition = prefetching_condition & (data["hyper-threads"] == 1)
    prefetching_max = data.loc[prefetching_condition, "GB / second"].max()
    
    return 100.0 / random_max * prefetching_max / 100.0

def get_latency(file_path):
    data = pd.read_csv(file_path)
    data = data.query('Type == "Server"')
    return data
    
    
def plot(vendors, latencies):
    fig, ax_left = plt.subplots()

    legend_handles = []

    # SWPF Improvement
    max_performance = 0
    ax_left.set_xlabel('Year', fontsize=15)
    ax_left.set_ylabel('Improvement by SW-Prefetching\n(over Random Access)', fontsize=15)
    ax_left.yaxis.set_major_formatter('{x:1.0f}x')
    ax_left.tick_params(axis='both', which='major', labelsize=14)

    for name, vendor in vendors.items():
        architectures = vendor[0]
        color = vendor[1]
        marker = vendor[2]

        years = [t[0] for t in architectures]
        names = [t[1] for t in architectures]
        performance = [t[2] for t in architectures]
        max_performance = max(max(performance), max_performance)

        handle, = ax_left.plot(years, performance, marker=marker, linestyle=':', color=color, linewidth=3, markersize=8, label=name)
        legend_handles.append(handle)  # Creates a line plot with markers


        label_offset = [t[3] for t in architectures]
        for i, value in enumerate(performance):
            plt.text(years[i], value + (max_performance * label_offset[i]), names[i], ha='center', va='bottom', fontsize=12)

    ax_left.set_ylim([1,max_performance * 1.1])
    ax_left.set_xlim(2011,2024)

    # Latencies
    if latencies:
        latency_color = '#81b29a'
        ax_right = ax_left.twinx()
        ax_right.set_ylabel('Latency (ns)')
        ax_right.set_ylim([0,500])
        ax_right.yaxis.label.set_color(latency_color)
        ax_right.tick_params(axis='y', colors=latency_color)
        handle, = ax_right.plot(latencies['Year'], latencies['Latency (ns)'], linestyle=':', linewidth=2, marker='o', color=latency_color, label='Latency')
        legend_handles.append(handle)

    ax_left.grid(True, color='#cccccc')

    ax_left.legend(loc='upper left', handles=legend_handles, ncol=3, fontsize=14)
    ax_left.spines[['right', 'top']].set_visible(False)

    #ax_left.legend()
    plt.tight_layout()
    plt.savefig('plots/trend.pdf')
    print('Created plots/trend.pdf')


cache_lines = 1

intel = [ 
    (2012, 'Sandy Bridge', get_prefetching_improvement('plots/Intel/sandy-bridge/out.csv', cache_lines), .01),
    (2013, 'Haswell', get_prefetching_improvement('plots/Intel/haswell/out.csv', cache_lines), .015),
    (2015, 'Skylake', get_prefetching_improvement('plots/Intel/skylake/out.csv', cache_lines), .015),
    (2019, 'Cascade Lake', get_prefetching_improvement('plots/Intel/cascade-lake-arnor/out.csv', cache_lines), -.05),
    (2023, 'Sapphire Rapids', get_prefetching_improvement('plots/Intel/sapphire-rapids/out.csv', cache_lines), .015)
]

arm = [
    (2012, 'Cortex-A53', get_prefetching_improvement('plots/ARM/cortex-a53/out.csv', cache_lines), -.05),
    (2015, 'Cortex-A72', get_prefetching_improvement('plots/ARM/cortex-a72/out.csv', cache_lines), .015),
    (2016, 'Cortex-A73', get_prefetching_improvement('plots/ARM/cortex-a73/out.csv', cache_lines), -.05),
    (2018, 'Cortex-A76', get_prefetching_improvement('plots/ARM/cortex-a76/out.csv', cache_lines), .015),
#    (2021, 'M1 Icestorm', get_prefetching_improvement('plots/ARM/apple-m1-little/out.csv', cache_lines), -.058),
#    (2021, 'M1 Firestorm', get_prefetching_improvement('plots/ARM/apple-m1-big/out.csv', cache_lines), .01),
]


amd = [
    (2017, 'Zen 1', get_prefetching_improvement('plots/AMD/zen1-desktop/out.csv', cache_lines), .02),
    (2019, 'Zen 2', get_prefetching_improvement('plots/AMD/zen2-desktop/out.csv', cache_lines), .02),
    (2022, 'Zen 4', get_prefetching_improvement('plots/AMD/zen4-desktop/out.csv', cache_lines), .015),
]

latency = None #get_latency('plots/latency.csv')

plot({ 'Intel': (intel, '#3d405b', 'v'), 'ARM': (arm, '#e07a5f', 's'), 'AMD': (amd, '#a01a2f', 'o')}, latency)

