#pragma once

#include <cstdint>
#include <optional>
#include "data.h"

namespace swprf
{
class SamplingConfig
{
public:
    enum Type
    {
        None,
        Address,
        Counter
    };

    SamplingConfig(const Type type, std::optional<std::string> counter_name, const std::uint64_t period, const std::uint8_t precise_ip) noexcept
            : _type(type), _counter_name(std::move(counter_name)), _period(period), _precise_ip(precise_ip)
    {
    }

    ~SamplingConfig() noexcept = default;

    [[nodiscard]] Type type() const noexcept { return _type; }
    [[nodiscard]] const std::optional<std::string>& counter_name() const noexcept { return _counter_name; }
    [[nodiscard]] std::uint64_t period() const noexcept { return _period; }
    [[nodiscard]] std::uint8_t precise_ip() const noexcept { return _precise_ip; }
private:
    Type _type;
    std::optional<std::string> _counter_name;
    std::uint64_t _period;
    std::uint8_t _precise_ip;
};

class BenchmarkConfig
{
public:
    BenchmarkConfig(std::uint32_t size_in_gb,
                    std::string page_size,
                    bool is_random,
                    std::vector<std::uint16_t>&& cores,
                    std::uint16_t prefetch_distance,
                    std::uint8_t prefetch_locality,
                    SamplingConfig&& sample_config,
                    std::optional<std::string> perf_profile_file_name,
                    std::optional<std::string> results_file_name,
                    bool is_results_per_cacheline,
                    bool is_interleaving)
            : _size_in_gb(size_in_gb),
              _allocation_strategy(swprf::AllocationStrategy::K4_Normal),
              _is_random(is_random),
              _cores(std::move(cores)),
              _prefetch_distance(prefetch_distance),
              _prefetch_locality(prefetch_locality),
              _sample_config(std::move(sample_config)),
              _perf_profile_file_name(std::move(perf_profile_file_name)),
              _results_file_name(std::move(results_file_name)),
              _is_results_per_cacheline(is_results_per_cacheline),
              _is_interleaving(is_interleaving)
    {
        if (page_size == "1G")
        {
            _allocation_strategy = swprf::AllocationStrategy::G1_HugePages;
        }
        else if (page_size == "2M")
        {
            _allocation_strategy = swprf::AllocationStrategy::M2_HugePages;
        }
        else if (page_size != "4K")
        {
            throw std::runtime_error{"Unknown page size (4K, 2M, or 1G allowed)."};
        }
    }

    ~BenchmarkConfig() = default;

    [[nodiscard]] std::uint32_t size_in_gb() const { return _size_in_gb; }
    [[nodiscard]] AllocationStrategy allocation_strategy() const { return _allocation_strategy; }
    [[nodiscard]] bool is_random() const { return _is_random; }
    [[nodiscard]] const std::vector<std::uint16_t>& cores() const { return _cores; }
    [[nodiscard]] std::uint16_t prefetch_distance() const { return _prefetch_distance; }
    [[nodiscard]] std::uint8_t prefetch_locality() const { return _prefetch_locality; }
    [[nodiscard]] const SamplingConfig& sample_config() const { return _sample_config; }
    [[nodiscard]] std::optional<std::string> perf_profile_file_name() const { return _perf_profile_file_name; }
    [[nodiscard]] std::optional<std::string> results_file_name() const { return _results_file_name; }
    [[nodiscard]] bool is_results_per_cacheline() const { return _is_results_per_cacheline; }
    [[nodiscard]] bool is_interleaving() const { return _is_interleaving; }

private:
    std::uint32_t _size_in_gb;
    AllocationStrategy _allocation_strategy;
    bool _is_random;
    std::vector<std::uint16_t> _cores;
    std::uint16_t _prefetch_distance;
    std::uint8_t _prefetch_locality;
    SamplingConfig _sample_config;
    std::optional<std::string> _perf_profile_file_name;
    std::optional<std::string> _results_file_name;
    bool _is_results_per_cacheline;
    bool _is_interleaving;
};
}