#pragma once

#include <cstdint>
#include <vector>

namespace swprf
{
class Compute
{
public:
    [[nodiscard]] static inline std::uint32_t murmur_32_scramble(std::uint32_t k)
    {
        /// Murmur3 32bit https://en.wikipedia.org/wiki/MurmurHash
        k *= 0xcc9e2d51;
        k = (k << 15) | (k >> 17);
        k *= 0x1b873593;
        return k;
    }

    [[nodiscard]] static inline std::uint32_t murmur_32(std::uint32_t val)
    {
        /// Murmur3 32bit https://en.wikipedia.org/wiki/MurmurHash
        auto h = 19553029U;

        /// Scramble
        h ^= murmur_32_scramble(val);
        h = (h << 13U) | (h >> 19U);
        h = h * 5U + 0xe6546b64;

        /// Finalize
        h ^= sizeof(std::uint32_t);
        h ^= h >> 16U;
        h *= 0x85ebca6b;
        h ^= h >> 13U;
        h *= 0xc2b2ae35;
        h ^= h >> 16U;

        return h;
    }

    [[nodiscard]] static inline std::uint32_t work(std::uint32_t val)
    {
        return murmur_32(val) + murmur_32(val >> 2U);
    }

        [[nodiscard]] static inline std::uint32_t heavy_work(std::uint32_t h)
    {
        /// Murmur https://www.sderosiaux.com/articles/2017/08/26/the-murmur3-hash-function--hashtables-bloom-filters-hyperloglog/
        h ^= h >> 16;
        h *= 0x85ebca6b;
        h ^= h >> 13;
        h *= 0xc2b2ae35;
        h ^= h >> 16;

        h = (h ^ (h >> 30)) / 0xbf58476d1ce4e5b9L;
        h = (h ^ (h >> 27)) * 0x94d049bb133111ebL;
        h = h ^ (h >> 31);

        h ^= h >> 16;
        h *= 0x85ebca6b;
        h ^= h >> 13;
        h *= 0xc2b2ae35;
        h ^= h >> 16;

        h = (h ^ (h >> 10)) / 0xbf58476d1ce521b9L;
        h = (h ^ (h >> 17)) * 0x94d049bb1331e2ebL;
        h = h ^ (h >> 13);

        h ^= h >> 12;
        h *= 0x85ebca6b;
        h ^= h >> 19;
        h *= 0xc5b6ae45;
        h ^= h >> 26;
        
        h = (h ^ (h >> 20)) / 0xbf58476d1ce4e5b9L;
        h = (h ^ (h >> 7)) * 0xf62839bb133111ebL;
        h = h ^ (h >> 19);

        h ^= h >> 6;
        h *= 0x22ebca6b;
        h ^= h >> 23;
        h *= 0x72b2ae35;
        h ^= h >> 14;

        return h;
    }
};

template<typename K, typename V>
class SimpleHashTable 
{
public:
    explicit SimpleHashTable(const size_t size) : _size(size)
    {
        _table.reserve(_size);
        auto random_engine = std::default_random_engine{64988430769U};
        for (auto i = 0U; i < _size; ++i) 
        {
            _table[i].key = random_engine();
            _table[i].value = random_engine();
        }
    }
    SimpleHashTable(SimpleHashTable&&) noexcept = default;
    ~SimpleHashTable() = default;

    void insert(const K& key, const V& value) 
    {
        auto idx = hash(key);
        _table[idx].key = key;
        _table[idx].value = value;
    }

    V get(const K& key) const 
    {
        return _table[hash(key)].value;
    }

private:
    size_t hash(const K& key) const 
    {
        return key % (_size - 1); // Modulo to fit within array size
    }

    struct KVPair 
    {
        K key;
        V value;
    };

    size_t _size;
    std::vector<KVPair> _table;
};
}
