#pragma once

#include <cstdint>
#include <vector>
#include <random>
#include <algorithm>
#include <cstdlib>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <iostream>

#define MAP_HUGE_2MB    (21 << MAP_HUGE_SHIFT)
#define MAP_HUGE_1GB    (30 << MAP_HUGE_SHIFT)

namespace swprf
{
enum AllocationStrategy
{
    K4_Normal,
    M2_HugePages,
    G1_HugePages
};

template<typename T>
class AlignedAllocator
{
public:
    using value_type = T;

    explicit AlignedAllocator(const AllocationStrategy strategy) : _strategy(strategy)
    {
    }

    AlignedAllocator(const AlignedAllocator&) = default;
    AlignedAllocator(AlignedAllocator&&) = default;

    AlignedAllocator& operator=(const AlignedAllocator&) noexcept = default;
    AlignedAllocator& operator=(AlignedAllocator&&) noexcept = default;

    T* allocate(const std::size_t count)
    {
        const auto size = sizeof(T) * count;

        if (_strategy == M2_HugePages)
        {
            return static_cast<T*> (::mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_HUGETLB | MAP_HUGE_2MB, -1, 0));
        }
        else if(_strategy == G1_HugePages)
        {
            return static_cast<T*> (::mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_HUGETLB | MAP_HUGE_1GB, -1, 0));
        }

        return static_cast<T*>(std::aligned_alloc(4096U, size));
    }

    void deallocate(T* ptr, const std::size_t count)
    {
        if (_strategy == M2_HugePages || _strategy == G1_HugePages)
        {
            ::munmap(ptr, sizeof(T) * count);
        }
        else
        {
            std::free(ptr);
        }
    }

private:
    AllocationStrategy _strategy = K4_Normal;
};

template<typename T>
class Workload
{
public:
    explicit Workload(std::size_t size, AllocationStrategy allocation_strategy);
    ~Workload() = default;

    Workload(Workload&&) noexcept = default;
    Workload& operator=(Workload&&) noexcept = default;

    void shuffle();

    [[nodiscard]] std::size_t size() const noexcept { return _access_pattern.size(); }
    [[nodiscard]] const std::vector<std::size_t>& access_pattern() const noexcept { return _access_pattern; }
    [[nodiscard]] const T& operator[](const std::size_t index) const noexcept { return _data[index]; }

    [[nodiscard]] std::pair<std::uintptr_t, std::uintptr_t> address_range() const noexcept { const auto start = std::uintptr_t(_data.data()); return std::make_pair(start, start + (_data.size() * sizeof(T))); }
private:
    std::vector<std::size_t> _access_pattern;
    std::vector<T, AlignedAllocator<T>> _data;
};

template<typename T>
Workload<T>::Workload(const std::size_t size, const AllocationStrategy allocation_strategy) : _data(AlignedAllocator<T>{allocation_strategy})
{
    /// Generate access access_pattern
    _access_pattern.resize(size);
    std::iota(_access_pattern.begin(), _access_pattern.end(), 0U);

    /// Generate data
    _data.resize(size);
    for (auto i = 0U; i < size; ++i)
    {
        _data[i] = i;
    }
}

template<typename T>
void Workload<T>::shuffle()
{
    auto random_engine = std::default_random_engine{64988430769U};
    std::shuffle(_access_pattern.begin(), _access_pattern.end(), random_engine);
}
}