#include <cstdint>
#include "data.h"
#include <argparse/argparse.hpp>
#include <nlohmann/json.hpp>
#include <perfcpp/event_counter.h>
#include <perfcpp/sampler.h>
#include <iostream>
#include <chrono>
#include <array>
#include <fstream>
#include <atomic>
#include <thread>
#include <exception>
#include "thread.h"
#include "compute.h"
#include "cache_lines.h"
#include "benchmark_config.h"


template<class T, std::uint32_t L>
__attribute__((noinline)) std::uint64_t exec_with_prefetching(const std::size_t from, const std::size_t to, const swprf::Workload<T>& workload, const std::uint16_t prefetch_distance)
{
    const auto& access_pattern = workload.access_pattern();

    auto value = 0ULL;
    for (auto i = from; i < to; ++i)
    {
        const auto index_to_prefetch = access_pattern[i + prefetch_distance];
        auto *cache_line_to_prefetch = &workload[index_to_prefetch];
        T::template prefetch<L>(cache_line_to_prefetch);

        const auto index_to_access = access_pattern[i];
        const auto& cache_line_to_access = workload[index_to_access];
        value += cache_line_to_access.compute();
    }

    return value;
}

template<class T, std::uint32_t L>
__attribute__((noinline)) std::uint64_t exec_with_interleaved_prefetching(const std::size_t from, const std::size_t to, const swprf::Workload<T>& workload, const std::uint16_t prefetch_distance)
{
    const auto& access_pattern = workload.access_pattern();

    auto value = 0ULL;
    for (auto i = from; i < to; ++i)
    {
        const auto index_to_prefetch = access_pattern[i + prefetch_distance];
        auto *cache_line_to_prefetch = &workload[index_to_prefetch];

        const auto index_to_access = access_pattern[i];
        const auto& cache_line_to_access = workload[index_to_access];

        /// Compute and prefetch interleaved.
        value += cache_line_to_access.template compute_and_prefetch_interleaved<L>(cache_line_to_prefetch);
    }

    return value;
}

template<class T>
__attribute__((noinline)) std::uint64_t exec_without_prefetching(const std::size_t from, const std::size_t to, const swprf::Workload<T>& workload)
{
    const auto& access_pattern = workload.access_pattern();

    auto value = 0ULL;
    for (auto i = from; i < to; ++i)
    {
        const auto index_to_access = access_pattern[i];
        const auto& cache_line_to_access = workload[index_to_access];
        value += cache_line_to_access.compute();
    }

    return value;
}

template<class T>
void benchmark(swprf::BenchmarkConfig&& config)
{
    /// Create the workload.
    const auto number_of_objects = (1024UL * 1024UL * 1024UL * config.size_in_gb()) / sizeof(T);

    std::cout << "Building workload for " << config.size_in_gb() << " GB (" << number_of_objects << " x " << typeid(T).name() << ")" << std::endl;
    auto workload = swprf::Workload<T>{number_of_objects, config.allocation_strategy()};

    if (config.is_random())
    {
        workload.shuffle();
    }

    std::cout << "Cores:";
    for (const auto core : config.cores())
    {
        std::cout << " " << core;
    }
    std::cout << std::endl;

    /// Check if prefetching is enabled (indicated by distance > 0).
    const auto is_prefetching = config.prefetch_distance() > 0U;

    /// Parse Perf profile.
    auto counter_list = std::vector<std::string>{};
    auto counter_definition_file = std::string{};
    if (config.perf_profile_file_name().has_value())
    {
        auto perf_profile_file = std::ifstream {config.perf_profile_file_name().value()};
        auto perf_profile = nlohmann::json::parse(perf_profile_file);
        if (perf_profile.contains("definitions"))
        {
            counter_definition_file = perf_profile["definitions"].get<std::string>();
        }

        if (perf_profile.contains("counter"))
        {
            for (const auto& counter : perf_profile["counter"])
            {
                counter_list.emplace_back(counter.get<std::string>());
            }
        }

    }

    /// Init Perf.
    auto counter_definitions = perf::CounterDefinition{std::move(counter_definition_file)};
    counter_definitions.add("ibs_op", perf::CounterConfig{11, 0}); // (1ULL<<19)

    auto event_counter = std::optional<perf::EventCounter>{std::nullopt};

    if (config.sample_config().type() == swprf::SamplingConfig::Type::None)
    {
        auto perf_config = perf::Config{};
        perf_config.include_child_threads(true);
        perf_config.include_kernel(false);
        perf_config.include_idle(false);
        event_counter.emplace(perf::EventCounter{counter_definitions, perf_config});
        event_counter->add(std::move(counter_list));
    }
    else
    {
        if (config.sample_config().counter_name().has_value())
        {
            /// User provided a specific counter to record.
            counter_list.clear();
            counter_list.push_back(config.sample_config().counter_name().value());
        }
        else
        {
            if (config.sample_config().type() == swprf::SamplingConfig::Type::Address)
            {
                /// For address keep only the first counter.
                counter_list.resize(1U);
            }
            else if (config.sample_config().type() == swprf::SamplingConfig::Type::Counter)
            {
                /// For counters keep only the first five counters.
                if (counter_list.size() > 5U)
                {
                    counter_list.resize(5U);
                }
            }
        }
    }


    /// Threads.
    auto start_barrier = std::atomic<bool>{false}; /// All threads will wait for this to be true.
    auto threads = std::vector<std::thread>{};
    threads.reserve(config.cores().size());

    const auto cache_lines_per_thread = workload.size() / config.cores().size();

    /// Samplers per core.
    auto *local_samplers = reinterpret_cast<perf::Sampler*>(std::malloc(sizeof(perf::Sampler) * config.cores().size()));

    /// Start recording.
    if (event_counter.has_value())
    {
        if (!event_counter->start())
        {
            std::cerr << "Could not start performance counters." << std::endl;
        }
    }

    /// Spawn working threads.
    for (auto thread_index = 0U; thread_index < config.cores().size(); ++thread_index)
    {
        threads.emplace_back([&start_barrier, is_prefetching, &config,  &workload, thread_index, cache_lines_per_thread, &counter_list, &counter_definitions, &local_samplers] {
            const auto start = thread_index * cache_lines_per_thread;
            const auto end = std::min((thread_index + 1U) * cache_lines_per_thread, workload.size());

            if (config.sample_config().type() != swprf::SamplingConfig::Type::None)
            {
                auto perf_config = perf::SampleConfig{};
                perf_config.precise_ip(config.sample_config().precise_ip()); /// Should be made as parameter
                //perf_config.include_kernel(false);
                //perf_config.include_idle(false);
                perf_config.period(config.sample_config().period());

                /// "Normal" sampling might use perf::Sampler::Type::CounterValues;

                auto sample_type = std::uint64_t(perf::Sampler::Type::Time | perf::Sampler::Type::CPU);
                if (config.sample_config().type() == swprf::SamplingConfig::Type::Address)
                {
                    sample_type |= perf::Sampler::Type::DataSource | perf::Sampler::Type::Weight | perf::Sampler::Type::LogicalMemAddress;
                }
                else if (config.sample_config().type() == swprf::SamplingConfig::Type::Counter)
                {
                    sample_type |= perf::Sampler::CounterValues;
                }

                std::ignore = new (&local_samplers[thread_index]) perf::Sampler{
                    counter_definitions,
                    std::vector<std::string>{counter_list},
                    sample_type,
                    perf_config};
            }

            /// Wait to start.
            while(start_barrier.load() == false);

            if (config.sample_config().type() != swprf::SamplingConfig::Type::None)
            {
                if (!local_samplers[thread_index].start())
                {
                    std::cerr << "Could not start performance counter sampling: " << local_samplers[thread_index].last_error()  << std::endl;
                }
            }

            /// Run.
            std::uint64_t value;

            if (is_prefetching)
            {
                /// Interleaved prefetching.
                if (config.is_interleaving())
                {
                    if (config.prefetch_locality() == 0U)
                    {
                        value = exec_with_interleaved_prefetching<T, 0U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 1U)
                    {
                        value = exec_with_interleaved_prefetching<T, 1U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 2U)
                    {
                        value = exec_with_interleaved_prefetching<T, 2U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 3U)
                    {
                        value = exec_with_interleaved_prefetching<T, 3U>(start, end, workload, config.prefetch_distance());
                    }
                    else
                    {
                        throw std::runtime_error{"Prefetch Locality not supported."};
                    }
                }

                /// "Normal" prefetching.
                else
                {
                    if (config.prefetch_locality() == 0U)
                    {
                        value = exec_with_prefetching<T, 0U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 1U)
                    {
                        value = exec_with_prefetching<T, 1U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 2U)
                    {
                        value = exec_with_prefetching<T, 2U>(start, end, workload, config.prefetch_distance());
                    }
                    else if (config.prefetch_locality() == 3U)
                    {
                        value = exec_with_prefetching<T, 3U>(start, end, workload, config.prefetch_distance());
                    }
                    else
                    {
                        throw std::runtime_error{"Prefetch Locality not supported."};
                    }
                }
            }

            /// No prefetching.
            else
            {
                value = exec_without_prefetching(start, end, workload);
            }

            asm volatile("" : "+r,m"(value) : : "memory");

            if (config.sample_config().type() != swprf::SamplingConfig::Type::None)
            {
                local_samplers[thread_index].stop();
            }
        });

        swprf::thread::pin(threads[thread_index], config.cores()[thread_index]);
    }

    /// Start time.
    const auto start = std::chrono::steady_clock::now();

    /// Let the threads go.
    start_barrier.store(true);

    /// Wait for threads.
    for (auto& thread : threads)
    {
        thread.join();
    }

    /// Stop time.
    const auto end = std::chrono::steady_clock::now();

    /// Stop recording counters.
    if (event_counter.has_value())
    {
        event_counter->stop();
    }

    /// Print the result.
    const auto number_of_cache_lines = config.is_results_per_cacheline() ? (workload.size() * (sizeof(T) / swprf::config::cache_line_size)) : workload.size();
    const auto ms = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();
    const auto ns = std::chrono::duration_cast<std::chrono::nanoseconds>(end - start).count();
    const auto ns_per_cache_line = double(ns) / number_of_cache_lines;
    const auto gb_per_second = config.size_in_gb() / (double(ms) / 1000U);

    /// Could be "cache line"  or "X cache lines"
    const auto measure_granularity = config.is_results_per_cacheline() ? "cache line" : (std::to_string(std::uint64_t{sizeof(T) / swprf::config::cache_line_size}) + " cache lines");

    std::cout << "Took " << ms << " ms (" << ns_per_cache_line << " ns / " << measure_granularity << ")." << std::endl;
    std::cout << "Bandwidth: " << gb_per_second << " GB/s" << std::endl;

    if (event_counter.has_value())
    {
        const auto result = event_counter->result(number_of_cache_lines);

        /// Store the results.
        if (config.results_file_name().has_value())
        {
            auto result_json = nlohmann::json{{"ns", ns_per_cache_line}, {"GB / second", gb_per_second}};
            for (const auto& [name, value] : result)
            {
                auto perf_counter_name = std::stringstream{};
                perf_counter_name << name << std::flush;
                result_json[perf_counter_name.str()] = value;
            }

            {
                auto results_file = std::ofstream{config.results_file_name().value()};
                results_file << result_json.dump(2) << std::flush;
            }
        }
        else
        {
            /// Print the performance counters.
            for (const auto& [name, value] : result)
            {
                std::cout << value  << " " << name << " / " << measure_granularity << "\n";
            }

            std::cout << std::flush;
        }
    }
    else if(config.sample_config().type() != swprf::SamplingConfig::Type::None)
    {
        /// Collect address and latency samples.
        if (config.sample_config().type() == swprf::SamplingConfig::Type::Address)
        {
            /// Combine samples from all cores.
            auto samples = std::vector<perf::Sample>{};
            for (auto i = 0U; i < config.cores().size(); ++i)
            {
                auto result = local_samplers[i].result();
                std::move(result.begin(), result.end(), std::back_inserter(samples));
                local_samplers[i].close();
            }

            /// Get the first timestamp.
            auto first_timestamp = std::numeric_limits<std::uint64_t>::max();
            for (const auto& sample : samples)
            {
                first_timestamp = std::min(sample.time().value_or(std::numeric_limits<std::uint64_t>::max()), first_timestamp);
            }

            {
                auto out_file = std::ofstream{config.results_file_name().value()};

                out_file
                        << "time,cpu,weight,is_user"
                        << ",is_load,is_store,is_prefetch"
                        << ",is_mem_hit,is_mem_miss,is_mem_na,is_mem_l1,is_mem_l2,is_mem_l3,is_mem_lfb,is_mem_ram,is_remote_cache,address"
                        << ",is_tlb_hit,is_tlb_miss,is_dtlb,is_stlb,is_tlb_walk,is_tlb_os_handler"
                        << ",is_workload,in_line_offset,in_block_offset"
                        ;

                /// Gather results.
                const auto data_range = workload.address_range();
                for (const auto& sample : samples)
                {
                    if (sample.time().has_value() && sample.cpu_id().has_value() && sample.weight().has_value() && sample.data_src().has_value() && sample.logical_memory_address().has_value())
                    {
                        const auto timestamp = sample.time().value() - first_timestamp;
                        const auto data_src = sample.data_src().value();

                        const auto is_in_workload = sample.logical_memory_address().value() >= std::get<0>(data_range) && sample.logical_memory_address().value() < std::get<1>(data_range);
                        const auto in_line_offset = sample.logical_memory_address().value() % 64U;
                        const auto in_block_offset = std::uint64_t( ((sample.logical_memory_address().value() - std::get<0>(data_range)) % sizeof(T)) / 64U);

                        out_file << "\n"
                                 << timestamp << "," << sample.cpu_id().value() << "," << sample.weight().value() << "," << static_cast<std::uint16_t>(sample.mode() == perf::Sample::Mode::User)
                                 << "," << static_cast<std::uint16_t>(data_src.is_load()) << "," << static_cast<std::uint16_t>(data_src.is_store()) << "," << static_cast<std::uint16_t>(data_src.is_prefetch())
                                 << "," << static_cast<std::uint16_t>(data_src.is_mem_hit()) << "," << static_cast<std::uint16_t>(data_src.is_mem_miss()) << "," << static_cast<std::uint16_t>(false) << "," << static_cast<std::uint16_t>(data_src.is_mem_l1()) << "," << static_cast<std::uint16_t>(data_src.is_mem_l2()) << "," << static_cast<std::uint16_t>(data_src.is_mem_l3()) << "," << static_cast<std::uint16_t>(data_src.is_mem_lfb()) << "," << static_cast<std::uint16_t>(data_src.is_mem_local_ram() || data_src.is_mem_remote_ram1() || data_src.is_mem_remote_ram2()) << "," << static_cast<std::uint16_t>(data_src.is_mem_remote_cce1() || data_src.is_mem_remote_cce2()) << "," << sample.logical_memory_address().value()
                                 << "," << static_cast<std::uint16_t>(data_src.is_tlb_hit()) << "," << static_cast<std::uint16_t>(data_src.is_tlb_miss()) << "," << static_cast<std::uint16_t>(data_src.is_tlb_l1()) << "," << static_cast<std::uint16_t>(data_src.is_tlb_l2()) << "," << static_cast<std::uint16_t>(data_src.is_tlb_walk()) << "," << static_cast<std::uint16_t>(false)
                                 << "," << static_cast<std::uint16_t>(is_in_workload) << "," << in_line_offset << "," << in_block_offset
                                ;
                    }
                }

                out_file << std::flush;
            }
        }

        /// Collect counter values
        else if (config.sample_config().type() == swprf::SamplingConfig::Type::Counter)
        {
            auto out_file = std::ofstream{config.results_file_name().value()};

            out_file << "time,cpu";
            for (const auto& counter_name : counter_list)
            {
                out_file << "," << counter_name;
            }

            for (auto core_id = 0U; core_id < config.cores().size(); ++core_id)
            {
                auto samples = local_samplers[core_id].result();

                if (samples.empty() == false)
                {
                    auto reference_counter_values = std::unordered_map<std::string, double>{};

                    auto reference_sample_id = 0U;
                    for (; reference_sample_id < samples.size(); ++reference_sample_id)
                    {
                        if (samples[reference_sample_id].counter_result().has_value())
                        {
                            const auto& reference_counter_result = samples[reference_sample_id].counter_result().value();
                            for (const auto& [name, value] : reference_counter_result)
                            {
                                reference_counter_values.insert(std::make_pair(name, value));
                            }
                            break;
                        }
                    }

                    for (auto sample_id = reference_sample_id + 1U; sample_id < samples.size(); ++sample_id)
                    {
                        const auto& sample = samples[sample_id];
                        if (sample.time().has_value() && sample.cpu_id().has_value())
                        {
                            out_file << "\n" << sample.time().value() << "," << sample.cpu_id().value();

                            for (const auto& counter_name : counter_list)
                            {
                                const auto value = sample.counter_result().value().get(counter_name).value();
                                const auto last_value = reference_counter_values[counter_name];

                                out_file << "," << (value - last_value);

                                reference_counter_values[counter_name] = value;
                            }
                        }
                    }
                }

                out_file << std::flush;
            }
        }
    }
}


int main(int count_arguments, char **arguments)
{
    static_assert(sizeof(cache_line) == swprf::config::cache_line_size && "'sizeof(cache_line)' and 'swprf::config::cache_line_size' do not match");

    auto argument_parser = argparse::ArgumentParser{"swpf_benchmark"};
    argument_parser.add_argument("--gb")
            .help("Size of data to access in GB")
            .default_value(std::uint32_t(2U))
            .action([](const std::string &value) { return std::uint32_t(std::stoi(value)); });
    argument_parser.add_argument("-r", "--random")
            .help("If set, access will be random")
            .implicit_value(true)
            .default_value(false);
    argument_parser.add_argument("--cores")
            .help("Configuration defining the cores (can be file or string with cores, e.g., '0 1 2')")
            .default_value(std::string{"0"})
            .action([](const std::string &value) { return value; });
    argument_parser.add_argument("-pd", "--prefetch-distance")
            .help("Prefetch distance")
            .default_value(std::uint16_t(0U))
            .action([](const std::string &value) { return std::uint16_t(std::stoi(value)); });
    argument_parser.add_argument("-pl", "--prefetch-locality")
            .help("Prefetch locality (0 = NTA; 1 = T2; 2 = T1; 3 = T0)")
            .default_value(std::uint8_t(0U))
            .action([](const std::string &value) { return std::uint8_t(std::stoi(value)); });
    argument_parser.add_argument("--perf-profile")
            .help("Configuration defining the perf profile")
            .default_value(std::optional<std::string>{std::nullopt})
            .action([](const std::string &value) { return std::make_optional(value); });
    argument_parser.add_argument("--sample-addr")
            .help("If set, address and latency sampling is activated (counters defined in the perf file)")
            .implicit_value(true)
            .default_value(false);
    argument_parser.add_argument("--sample-values")
            .help("If set, sampling counter values is activated (counters defined in the perf file)")
            .implicit_value(true)
            .default_value(false);

    /*
     * TODO: Enable if we can sample more than addresses.
     * argument_parser.add_argument("--sample-addr")
     *       .help("If set (only together with sampling), address sampling is used; address-supporting counter must be first in perf file list.")
     *       .implicit_value(true)
     *       .default_value(false);
     **/
    argument_parser.add_argument("--sample-counter")
            .help("Name of the counter to sample. If not set, will use the counters from profile.")
            .default_value(std::optional<std::string>{std::nullopt})
            .action([](const std::string &value) { return std::make_optional(value); });
    argument_parser.add_argument("--sample-precise-ip")
            .help("Value for precise_ip, used for sampling.")
            .default_value(std::uint16_t{3U})
            .action([](const std::string &value) {  return std::uint16_t(std::stoi(value)); });
    argument_parser.add_argument("--sample-period")
            .help("Value for sample period.")
            .default_value(std::uint64_t{100U})
            .action([](const std::string &value) {  return std::uint64_t(std::stoi(value)); });
    argument_parser.add_argument("-cl", "--cache-lines")
            .help("Number of cache lines accessed together.")
            .default_value(std::uint16_t(1U))
            .action([](const std::string &value) { return std::uint16_t(std::stoi(value)); });
    argument_parser.add_argument("--prefetch-lines")
            .help("Number of cache lines prefetched together.")
            .default_value(std::uint16_t(0U))
            .action([](const std::string &value) { return std::uint16_t(std::stoi(value)); });
    argument_parser.add_argument("--adj", "--adjacent")
            .help("If set, execute adjacent instead of squential prefetching (only every 2nd cache line).")
            .implicit_value(true)
            .default_value(false);
    argument_parser.add_argument("--interleaved")
            .help("If set, use interleaved prefetching")
            .implicit_value(true)
            .default_value(false);
    argument_parser.add_argument("--page-size")
            .help("Size of allocated memory pages (4K, 2M, 1G).")
            .default_value(std::string{"4K"});
    argument_parser.add_argument("--result-per-cacheline")
            .help("Relative results to number of cache lines?")
            .implicit_value(true)
            .default_value(false);
    argument_parser.add_argument("-o", "--out")
            .help("File to store the results.")
            .default_value(std::optional<std::string>{std::nullopt})
            .action([](const std::string &value) { return std::make_optional(value); });

    /// Parse arguments.
    try
    {
        argument_parser.parse_args(count_arguments, arguments);
    }
    catch (std::runtime_error &e)
    {
        std::cout << argument_parser << std::endl;
        return 1;
    }

    /// Parse cores.
    auto cores_text = argument_parser.get<std::string>("--cores");
    {   /// Try to read as a file; read first line into cores string.
        auto cores_file_stream = std::ifstream{cores_text};
        if (cores_file_stream.good())
        {
            std::getline(cores_file_stream, cores_text);
        }
    }
    auto cores = std::vector<std::uint16_t>{};
    auto cores_stream = std::istringstream{cores_text};
    {
        std::string core;
        while(std::getline(cores_stream, core, ' '))
        {
            cores.emplace_back(std::stoul(core));
        }
    }

    const auto item_size = argument_parser.get<std::uint16_t>("--cache-lines");
    auto prefetch_size = argument_parser.get<std::uint16_t>("--prefetch-lines");
    if (prefetch_size == 0)
    {
        prefetch_size = item_size;
    }
    const auto is_adjacent_prefetching = argument_parser.get<bool>("--adjacent");

    /// Sampling type
    auto sampling_type = swprf::SamplingConfig::Type::None;
    if (argument_parser.get<bool>("--sample-addr"))
    {
        sampling_type = swprf::SamplingConfig::Type::Address;
    }
    else if (argument_parser.get<bool>("--sample-values"))
    {
        sampling_type = swprf::SamplingConfig::Type::Counter;
    }

    /// Benchmark config
    auto config = swprf::BenchmarkConfig{
        argument_parser.get<std::uint32_t>("--gb"),
        argument_parser.get<std::string>("--page-size"),
        argument_parser.get<bool>("--random"),
        std::move(cores),
        argument_parser.get<std::uint16_t>("-pd"),
        argument_parser.get<std::uint8_t>("-pl"),
        swprf::SamplingConfig{
            sampling_type,
            argument_parser.get<std::optional<std::string>>("--sample-counter"),
            argument_parser.get<std::uint64_t>("--sample-period"),
            static_cast<std::uint8_t>(argument_parser.get<std::uint16_t>("--sample-precise-ip"))
        },
        argument_parser.get<std::optional<std::string>>("--perf-profile"),
        argument_parser.get<std::optional<std::string>>("--out"),
        argument_parser.get<bool>("--result-per-cacheline") || item_size == 1U,
        argument_parser.get<bool>("--interleaved")
    };

    /****
     *
     * Sequential Prefetch Pattern: Prefetch some or all cache lines sequentially
     *
    ****/
    if (is_adjacent_prefetching == false)
    {
        /****
         * Single Cache Line
         ****/
        if (item_size == 1U)
        {
            benchmark<cache_lines_1_pref_1_seq>(std::move(config));
        }
        /****
         * 2 Cache Lines
         ****/
        else if (item_size == 2U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_2_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_2_pref_2_seq>(std::move(config));
            }
        }
        /****
         * 4 Cache Lines
         ****/
        else if (item_size == 4U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_4_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_4_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_4_pref_4_seq>(std::move(config));
            }
        }
        /****
         * 6 Cache Lines
         ****/
        else if (item_size == 6U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_6_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_6_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_6_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_6_pref_6_seq>(std::move(config));
            }
        }
        /****
         * 8 Cache Lines
         ****/
        else if (item_size == 8U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_8_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_8_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_8_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_8_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_8_pref_8_seq>(std::move(config));
            }
        }
        /****
         * 10 Cache Lines
         ****/
        else if (item_size == 10U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_10_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_10_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_10_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_10_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_10_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_10_pref_10_seq>(std::move(config));
            }
        }
        /****
         * 12 Cache Lines
         ****/
        else if (item_size == 12U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_12_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_12_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_12_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_12_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_12_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_12_pref_10_seq>(std::move(config));
            }
            else if (prefetch_size == 12U)
            {
                benchmark<cache_lines_12_pref_12_seq>(std::move(config));
            }
        }
        /****
         * 14 Cache Lines
         ****/
        else if (item_size == 14U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_14_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_14_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_14_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_14_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_14_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_14_pref_10_seq>(std::move(config));
            }
            else if (prefetch_size == 12U)
            {
                benchmark<cache_lines_14_pref_12_seq>(std::move(config));
            }
            else if (prefetch_size == 14U)
            {
                benchmark<cache_lines_14_pref_14_seq>(std::move(config));
            }
        }
        /****
         * 16 Cache Lines
         ****/
        else if (item_size == 16U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_16_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_16_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_16_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_16_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_16_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_16_pref_10_seq>(std::move(config));
            }
            else if (prefetch_size == 12U)
            {
                benchmark<cache_lines_16_pref_12_seq>(std::move(config));
            }
            else if (prefetch_size == 14U)
            {
                benchmark<cache_lines_16_pref_14_seq>(std::move(config));
            }
            else if (prefetch_size == 16U)
            {
                benchmark<cache_lines_16_pref_16_seq>(std::move(config));
            }
        }
        /****
         * 24 Cache Lines
         ****/
        else if (item_size == 24U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_24_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_24_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_24_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_24_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_24_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_24_pref_10_seq>(std::move(config));
            }
            else if (prefetch_size == 12U)
            {
                benchmark<cache_lines_24_pref_12_seq>(std::move(config));
            }
            else if (prefetch_size == 14U)
            {
                benchmark<cache_lines_24_pref_14_seq>(std::move(config));
            }
            else if (prefetch_size == 16U)
            {
                benchmark<cache_lines_24_pref_16_seq>(std::move(config));
            }
            else if (prefetch_size == 18U)
            {
                benchmark<cache_lines_24_pref_18_seq>(std::move(config));
            }
            else if (prefetch_size == 20U)
            {
                benchmark<cache_lines_24_pref_20_seq>(std::move(config));
            }
            else if (prefetch_size == 22U)
            {
                benchmark<cache_lines_24_pref_22_seq>(std::move(config));
            }
            else if (prefetch_size == 24U)
            {
                benchmark<cache_lines_24_pref_24_seq>(std::move(config));
            }
        }
        /****
         * 32 Cache Lines
         ****/
        else if (item_size == 32U)
        {
            if (prefetch_size == 1U)
            {
                benchmark<cache_lines_32_pref_1_seq>(std::move(config));
            }
            else if (prefetch_size == 2U)
            {
                benchmark<cache_lines_32_pref_2_seq>(std::move(config));
            }
            else if (prefetch_size == 4U)
            {
                benchmark<cache_lines_32_pref_4_seq>(std::move(config));
            }
            else if (prefetch_size == 6U)
            {
                benchmark<cache_lines_32_pref_6_seq>(std::move(config));
            }
            else if (prefetch_size == 8U)
            {
                benchmark<cache_lines_32_pref_8_seq>(std::move(config));
            }
            else if (prefetch_size == 10U)
            {
                benchmark<cache_lines_32_pref_10_seq>(std::move(config));
            }
            else if (prefetch_size == 12U)
            {
                benchmark<cache_lines_32_pref_12_seq>(std::move(config));
            }
            else if (prefetch_size == 14U)
            {
                benchmark<cache_lines_32_pref_14_seq>(std::move(config));
            }
            else if (prefetch_size == 16U)
            {
                benchmark<cache_lines_32_pref_16_seq>(std::move(config));
            }
            else if (prefetch_size == 18U)
            {
                benchmark<cache_lines_32_pref_18_seq>(std::move(config));
            }
            else if (prefetch_size == 20U)
            {
                benchmark<cache_lines_32_pref_20_seq>(std::move(config));
            }
            else if (prefetch_size == 22U)
            {
                benchmark<cache_lines_32_pref_22_seq>(std::move(config));
            }
            else if (prefetch_size == 24U)
            {
                benchmark<cache_lines_32_pref_24_seq>(std::move(config));
            }
            else if (prefetch_size == 26U)
            {
                benchmark<cache_lines_32_pref_26_seq>(std::move(config));
            }
            else if (prefetch_size == 28U)
            {
                benchmark<cache_lines_32_pref_28_seq>(std::move(config));
            }
            else if (prefetch_size == 30U)
            {
                benchmark<cache_lines_32_pref_30_seq>(std::move(config));
            }
            else if (prefetch_size == 32U)
            {
                benchmark<cache_lines_32_pref_32_seq>(std::move(config));
            }
        }
        else
        {
            std::cout << item_size << " cache lines not supported" << std::endl;
            return 1;
        }
    }
    /****
     *
     * Adjacent Prefetch Pattern: Prefetch only every second cache line.
     *
    ****/
    else if (is_adjacent_prefetching && config.is_interleaving() == false)
    {
        /****
         * 2 Cache Lines
         ****/
        if (item_size == 2U)
        {
            benchmark<cache_lines_2_pref_1_adj>(std::move(config));
        }
        /****
         * 4 Cache Lines
         ****/
        else if (item_size == 4U)
        {
            benchmark<cache_lines_4_pref_2_adj>(std::move(config));
        }
        /****
         * 6 Cache Lines
         ****/
        else if (item_size == 6U)
        {
            benchmark<cache_lines_6_pref_3_adj>(std::move(config));
        }
        /****
         * 8 Cache Lines
         ****/
        else if (item_size == 8U)
        {
            benchmark<cache_lines_8_pref_4_adj>(std::move(config));
        }
        /****
         * 10 Cache Lines
         ****/
        else if (item_size == 10U)
        {
            benchmark<cache_lines_10_pref_5_adj>(std::move(config));
        }
        /****
         * 12 Cache Lines
         ****/
        else if (item_size == 12U)
        {
            benchmark<cache_lines_12_pref_6_adj>(std::move(config));
        }
        /****
         * 14 Cache Lines
         ****/
        else if (item_size == 14U)
        {
            benchmark<cache_lines_14_pref_7_adj>(std::move(config));
        }
        /****
         * 16 Cache Lines
         ****/
        else if (item_size == 16U)
        {
            benchmark<cache_lines_16_pref_8_adj>(std::move(config));
        }
        /****
         * 24 Cache Lines
         ****/
        else if (item_size == 24U)
        {
            benchmark<cache_lines_24_pref_12_adj>(std::move(config));
        }
        /****
         * 32 Cache Lines
         ****/
        else if (item_size == 32U)
        {
            benchmark<cache_lines_32_pref_16_adj>(std::move(config));
        }
        else
        {
            std::cout << item_size << " cache lines not supported" << std::endl;
            return 1;
        }
    }

    return 0;
}
